# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from gluonts.core.serde.flat import clone
from gluonts import time_feature, transform
from gluonts.core.component import equals
from gluonts.dataset.field_names import FieldName
from gluonts.mx import CanonicalRNNEstimator

from pandas.tseries.frequencies import to_offset


def test_map_transformation():
    tran = transform.VstackFeatures(
        output_field="dynamic_feat",
        input_fields=["age", "time_feat"],
        drop_inputs=True,
    )

    assert equals(tran, clone(tran))
    assert not equals(tran, clone(tran, {"drop_inputs": False}))


def test_add_time_features():
    tran = transform.AddTimeFeatures(
        start_field=FieldName.START,
        target_field=FieldName.TARGET,
        output_field="time_feat",
        time_features=[
            time_feature.day_of_week,
            time_feature.day_of_month,
            time_feature.month_of_year,
        ],
        pred_length=10,
    )

    tran2 = clone(
        tran,
        {
            "time_features": [
                time_feature.day_of_week,
                time_feature.day_of_month,
            ]
        },
    )

    assert equals(tran, clone(tran))
    assert not equals(tran, tran2)


def test_filter_transformation():
    prediction_length = 10
    tran1 = transform.FilterTransformation(
        lambda x: x["target"].shape[-1] > prediction_length
    )
    # serde.flat.clone(tran1) does not work on
    tran2 = transform.FilterTransformation(
        lambda x: x["target"].shape[-1] > prediction_length
    )
    tran3 = transform.FilterTransformation(
        condition=lambda x: x["target"].shape[-1] < prediction_length
    )

    assert equals(tran1, tran2)
    assert not equals(tran1, tran3)


def test_exp_num_sampler():
    sampler = transform.ExpectedNumInstanceSampler(num_instances=4)
    assert equals(sampler, clone(sampler))
    assert not equals(sampler, clone(sampler, {"num_instances": 5}))


def test_continuous_time_sampler():
    sampler = transform.ContinuousTimeUniformSampler(num_instances=4)
    assert equals(sampler, clone(sampler))
    assert not equals(sampler, clone(sampler, {"num_instances": 5}))


def test_instance_splitter():
    splitter = transform.InstanceSplitter(
        target_field=FieldName.TARGET,
        is_pad_field=FieldName.IS_PAD,
        start_field=FieldName.START,
        forecast_start_field=FieldName.FORECAST_START,
        instance_sampler=transform.ExpectedNumInstanceSampler(num_instances=4),
        past_length=100,
        future_length=10,
        time_series_fields=["dynamic_feat", "observed_values"],
    )

    splitter2 = clone(
        splitter,
        {
            "instance_sampler": transform.ExpectedNumInstanceSampler(
                num_instances=5
            )
        },
    )
    assert equals(splitter, clone(splitter))
    assert not equals(splitter, splitter2)


def test_continuous_time_splitter():
    splitter = transform.ContinuousTimeInstanceSplitter(
        past_interval_length=1,
        future_interval_length=1,
        instance_sampler=transform.ContinuousTimePointSampler(),
        freq=to_offset("H"),
    )

    splitter2 = transform.ContinuousTimeInstanceSplitter(
        past_interval_length=1,
        future_interval_length=1,
        instance_sampler=transform.ContinuousTimePointSampler(min_past=1.0),
        freq=to_offset("H"),
    )

    assert equals(splitter, clone(splitter))
    assert not equals(splitter, splitter2)


def test_chain():
    chain = transform.Chain(
        [
            transform.AddTimeFeatures(
                start_field=FieldName.START,
                target_field=FieldName.TARGET,
                output_field="time_feat",
                time_features=[
                    time_feature.day_of_week,
                    time_feature.day_of_month,
                    time_feature.month_of_year,
                ],
                pred_length=10,
            ),
            transform.AddAgeFeature(
                target_field=FieldName.TARGET,
                output_field="age",
                pred_length=10,
                log_scale=True,
            ),
            transform.AddObservedValuesIndicator(
                target_field=FieldName.TARGET, output_field="observed_values"
            ),
        ]
    )

    assert equals(chain, clone(chain))
    assert not equals(chain, clone(chain, {"transformations": []}))

    another_chain = transform.Chain(
        [
            transform.AddTimeFeatures(
                start_field=FieldName.START,
                target_field=FieldName.TARGET,
                output_field="time_feat",
                time_features=[
                    time_feature.day_of_week,
                    time_feature.day_of_month,
                    time_feature.month_of_year,
                ],
                pred_length=10,
            ),
            transform.AddAgeFeature(
                target_field=FieldName.TARGET,
                output_field="age",
                pred_length=10,
                log_scale=False,
            ),
            transform.AddObservedValuesIndicator(
                target_field=FieldName.TARGET, output_field="observed_values"
            ),
        ]
    )
    assert not equals(chain, another_chain)


def test_gluon_predictor():
    train_length = 100
    pred_length = 10

    estimator = CanonicalRNNEstimator("5min", train_length, pred_length)

    assert equals(estimator, clone(estimator))
    assert not equals(estimator, clone(estimator, {"freq": "1h"}))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Union, Optional

import numpy as np
from pydantic import BaseModel, root_validator


class FeatureImportanceResult(BaseModel):
    target: List[Union[List[float], float]]
    feat_static_cat: List[Union[List[float], float]]
    feat_static_real: List[Union[List[float], float]]
    past_feat_dynamic_real: List[Union[List[float], float]]
    feat_dynamic_real: List[Union[List[float], float]]
    feat_dynamic_cat: List[Union[List[float], float]]

    @root_validator()
    def check_shape(cls, values):
        """
        Validate the second dimension is the same for 2d results and all fields share the same dimensionality
        For example, time aligned results with dimension of (features, pred_length), the pred_length shall be the same
        :param values:
        :return:
        """
        dim = np.array(values.get("target")).ndim
        assert (
            0 < dim <= 2
        ), "expected the feature importances array to be in the dimension of 1d or 2d only but got {dim}d from target"
        for key, value in values.items():
            if value:
                assert (
                    np.array(value).ndim == dim
                ), f"dimension mismatch {key} with dim {np.array(value).ndim} and target with dim {dim} "
        if dim == 1:
            return values
        shape = np.shape(values.get("target"))[dim - 1]
        for key, value in values.items():
            if value:
                assert (
                    np.shape(value)[dim - 1] == shape
                ), f"shape mismatch {key} with shape {np.shape(value)} and target with shape {shape} "
        return values

    def mean(self, axis=None) -> "FeatureImportanceResult":
        mean_dict = {}
        for key, val in self.dict().items():
            mean_dict[key] = np.mean(val, axis=axis).tolist() if val else []
        return FeatureImportanceResult(**mean_dict)


class ExplanationResult(BaseModel):
    time_quantile_aggregated_result: Optional[FeatureImportanceResult]
    quantile_aggregated_result: FeatureImportanceResult



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import math
import sys

import mxnet as mx

import mxnet.ndarray as nd
import numpy as np
import pytest

from gluonts.mx.model.gp_forecaster.gaussian_process import GaussianProcess

from gluonts.mx.context import check_gpu_support
from gluonts.mx.kernels import RBFKernel
from gluonts.mx.linalg_util import jitter_cholesky, jitter_cholesky_eig


# This test verifies that both eigenvalue decomposition and iterative jitter method
# make a non-positive definite matrix positive definite to be able to compute the cholesky.
# Both gpu and cpu as well as single and double precision are tested.
@pytest.mark.skipif(
    sys.platform == "linux",
    reason="skipping since potrf crashes on mxnet 1.6.0 on linux when matrix is not spd",
)
@pytest.mark.parametrize("ctx", ["cpu", "gpu"])
@pytest.mark.parametrize("jitter_method", ["iter", "eig"])
@pytest.mark.parametrize("float_type", [np.float32, np.float64])
def test_jitter_unit(jitter_method, float_type, ctx) -> None:
    # TODO: Enable GPU tests on Jenkins
    if ctx == "gpu" and not check_gpu_support():
        return

    with mx.Context(ctx):
        matrix = nd.array(
            [[[1, 2], [3, 4]], [[10, 100], [-21.5, 41]]], dtype=float_type
        )
        F = mx.nd
        num_data_points = matrix.shape[1]
        if jitter_method == "eig":
            L = jitter_cholesky_eig(F, matrix, num_data_points, float_type)
        elif jitter_method == "iter":
            L = jitter_cholesky(F, matrix, num_data_points, float_type)

    assert np.sum(np.isnan(L.asnumpy())) == 0, "NaNs in Cholesky factor!"


# This test tests that the noiseless sample generated from the synthetic example does not have NaNs in it.
# Without the jitter method, NaNs occurs on the gpu for single and double precision and on the cpu for only single
# precision.  This test verifies that applying the default jitter method fixes these numerical issues on both cpu
# and gpu and for single and double precision.
@pytest.mark.skipif(
    sys.platform == "linux",
    reason="skipping since potrf crashes on mxnet 1.6.0 on linux when matrix is not spd",
)
@pytest.mark.parametrize("ctx", ["cpu", "gpu"])
@pytest.mark.parametrize("jitter_method", ["iter", "eig"])
@pytest.mark.parametrize("float_type", [np.float32, np.float64])
def test_jitter_synthetic_gp(jitter_method, float_type, ctx) -> None:
    # TODO: Enable GPU tests on Jenkins
    if ctx == "gpu" and not check_gpu_support():
        return
    # Initialize problem parameters
    batch_size = 1
    prediction_length = 50
    context_length = 5
    num_samples = 3

    with mx.Context(ctx):
        # Initialize test data to generate Gaussian Process from
        lb = -5
        ub = 5
        dx = (ub - lb) / (prediction_length - 1)
        x_test = nd.arange(lb, ub + dx, dx, dtype=float_type).reshape(-1, 1)
        x_test = nd.tile(x_test, reps=(batch_size, 1, 1))

        # Define the GP hyper parameters
        amplitude = nd.ones((batch_size, 1, 1), dtype=float_type)
        length_scale = math.sqrt(0.4) * nd.ones_like(amplitude)
        sigma = math.sqrt(1e-5) * nd.ones_like(amplitude)

        # Instantiate desired kernel object and compute kernel matrix
        rbf_kernel = RBFKernel(amplitude, length_scale)

        # Generate samples from 0 mean Gaussian process with RBF Kernel and plot it
        gp = GaussianProcess(
            sigma=sigma,
            kernel=rbf_kernel,
            prediction_length=prediction_length,
            context_length=context_length,
            num_samples=num_samples,
            float_type=float_type,
            jitter_method=jitter_method,
            sample_noise=False,  # Returns sample without noise
        )

        # Generate training set on subset of interval using the sine function
        x_train = nd.array([-4, -3, -2, -1, 1], dtype=float_type).reshape(
            context_length, 1
        )
        x_train = nd.tile(x_train, reps=(batch_size, 1, 1))
        y_train = nd.sin(x_train.squeeze(axis=2))

        # Predict exact GP using the GP predictive mean and covariance using the same fixed hyper-parameters
        samples, predictive_mean, predictive_std = gp.exact_inference(
            x_train, y_train, x_test
        )

    assert (
        np.sum(np.isnan(samples.asnumpy())) == 0
    ), "NaNs in predictive samples!"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Union, Type

import numpy as np
import pandas as pd

from gluonts.core.component import validated
from gluonts.dataset.common import DataEntry
from gluonts.dataset.field_names import FieldName
from gluonts.time_feature import TimeFeature

from ._base import MapTransformation, SimpleTransformation


def target_transformation_length(
    target: np.ndarray, pred_length: int, is_train: bool
) -> int:
    return target.shape[-1] + (0 if is_train else pred_length)


class MissingValueImputation:
    """
    The parent class for all the missing value imputation classes.

    You can just implement your own inheriting this class.
    """

    @validated()
    def __init__(self) -> None:
        pass

    def __call__(self, values: np.ndarray) -> np.ndarray:
        """

        Parameters
        ----------
        values : the array of values with or without nans

        Returns
        -------
        values : the array of values with the nans replaced according to the
          method used.

        """
        raise NotImplementedError()


class LeavesMissingValues(MissingValueImputation):
    """
    Just leaves the missing values untouched.
    """

    def __call__(self, values: np.ndarray) -> np.ndarray:
        return values


class DummyValueImputation(MissingValueImputation):
    """
    This class replaces all the missing values with the same dummy value given
    in advance.
    """

    @validated()
    def __init__(self, dummy_value: float = 0.0) -> None:
        self.dummy_value = dummy_value

    def __call__(self, values: np.ndarray) -> np.ndarray:
        nan_indices = np.where(np.isnan(values))
        values[nan_indices] = self.dummy_value
        return values


class MeanValueImputation(MissingValueImputation):
    """
    This class replaces all the missing values with the mean of the non missing
    values.

    Careful this is not a 'causal' method in the sense that it leaks
    information about the furture in the imputation. You may prefer to use
    CausalMeanValueImputation instead.
    """

    def __call__(self, values: np.ndarray) -> np.ndarray:
        if len(values) == 1 or np.isnan(values).all():
            return DummyValueImputation()(values)
        nan_indices = np.where(np.isnan(values))
        values[nan_indices] = np.nanmean(values)
        return values


class LastValueImputation(MissingValueImputation):
    """
    This class replaces each missing value with the last value that was not
    missing.

    (If the first values are missing, they are replaced by the closest non
    missing value.)
    """

    def __call__(self, values: np.ndarray) -> np.ndarray:
        if len(values) == 1 or np.isnan(values).all():
            return DummyValueImputation()(values)
        values = np.expand_dims(values, axis=0)

        mask = np.isnan(values)
        idx = np.where(~mask, np.arange(mask.shape[1]), 0)
        np.maximum.accumulate(idx, axis=1, out=idx)
        out = values[np.arange(idx.shape[0])[:, None], idx]

        values = np.squeeze(out)
        # in case we need to replace nan at the start of the array
        mask = np.isnan(values)
        values[mask] = np.interp(
            np.flatnonzero(mask), np.flatnonzero(~mask), values[~mask]
        )

        return values


class CausalMeanValueImputation(MissingValueImputation):
    """
    This class replaces each missing value with the average of all the values
    up to this point.

    (If the first values are missing, they are replaced by the closest non
    missing value.)
    """

    def __call__(self, values: np.ndarray) -> np.ndarray:
        if len(values) == 1 or np.isnan(values).all():
            return DummyValueImputation()(values)
        mask = np.isnan(values)

        # we cannot compute the mean with this method if there are nans so we
        # do a temporary fix of the nan just for the mean computation using
        # this:
        last_value_imputation = LastValueImputation()
        value_no_nans = last_value_imputation(values)

        # We do the cumulative sum shifted by one indices:
        adjusted_values_to_causality = np.concatenate(
            (np.repeat(0.0, 1), value_no_nans[:-1])
        )
        cumsum = np.cumsum(adjusted_values_to_causality)

        # We get the indices of the elements shifted by one indices:
        indices = np.linspace(0, len(value_no_nans) - 1, len(value_no_nans))

        ar_res = cumsum / indices.astype(float)
        values[mask] = ar_res[mask]

        # make sure that we do not leave the potential nan in the first
        # position:
        values[0] = value_no_nans[0]

        return values


class RollingMeanValueImputation(MissingValueImputation):
    """
    This class replaces each missing value with the average of all the last
    window_size (default=10) values.

    (If the first values are missing, they are replaced by the closest non
    missing value.)
    """

    @validated()
    def __init__(self, window_size: int = 10) -> None:
        self.window_size = 1 if window_size < 1 else window_size

    def __call__(self, values: np.ndarray) -> np.ndarray:
        if len(values) == 1 or np.isnan(values).all():
            return DummyValueImputation()(values)
        mask = np.isnan(values)

        # we cannot compute the mean with this method if there are nans so we
        # do a temporary fix of the nan just for the mean computation using
        # this:
        last_value_imputation = LastValueImputation()
        value_no_nans = last_value_imputation(values)

        adjusted_values_to_causality = np.concatenate(
            (
                np.repeat(value_no_nans[0], self.window_size + 1),
                value_no_nans[:-1],
            )
        )

        cumsum = np.cumsum(adjusted_values_to_causality)

        ar_res = (
            cumsum[self.window_size :] - cumsum[: -self.window_size]
        ) / float(self.window_size)

        values[mask] = ar_res[mask]

        # make sure that we do not leave the potential nan in the first
        # position:
        values[0] = value_no_nans[0]

        return values


class AddObservedValuesIndicator(SimpleTransformation):
    """
    Replaces missing values in a numpy array (NaNs) with a dummy value and adds
    an "observed"-indicator that is ``1`` when values are observed and ``0``
    when values are missing.

    Parameters
    ----------
    target_field
        Field for which missing values will be replaced
    output_field
        Field name to use for the indicator
    imputation_method
        One of the methods from ImputationStrategy. If set to None, no
        imputation is done and only the indicator is included.
    """

    @validated()
    def __init__(
        self,
        target_field: str,
        output_field: str,
        imputation_method: Optional[
            MissingValueImputation
        ] = DummyValueImputation(0.0),
        dtype: Type = np.float32,
    ) -> None:
        self.target_field = target_field
        self.output_field = output_field
        self.dtype = dtype
        self.imputation_method = imputation_method

    def transform(self, data: DataEntry) -> DataEntry:
        value = data[self.target_field]
        nan_entries = np.isnan(value)

        if self.imputation_method is not None:
            if nan_entries.any():
                value = value.copy()
                data[self.target_field] = self.imputation_method(value)

        data[self.output_field] = np.invert(
            nan_entries, out=nan_entries
        ).astype(self.dtype, copy=False)
        return data


class AddConstFeature(MapTransformation):
    """
    Expands a `const` value along the time axis as a dynamic feature, where the
    T-dimension is defined as the sum of the `pred_length` parameter and the
    length of a time series specified by the `target_field`.

    If `is_train=True` the feature matrix has the same length as the `target`
    field. If `is_train=False` the feature matrix has length
    `len(target) + pred_length`.

    Parameters
    ----------
    output_field
        Field name for output.
    target_field
        Field containing the target array. The length of this array will be
        used.
    pred_length
        Prediction length (this is necessary since features have to be
        available in the future)
    const
        Constant value to use.
    dtype
        Numpy dtype to use for resulting array.
    """

    @validated()
    def __init__(
        self,
        output_field: str,
        target_field: str,
        pred_length: int,
        const: float = 1.0,
        dtype: Type = np.float32,
    ) -> None:
        self.pred_length = pred_length
        self.const = const
        self.dtype = dtype
        self.output_field = output_field
        self.target_field = target_field

    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        length = target_transformation_length(
            data[self.target_field], self.pred_length, is_train=is_train
        )
        data[self.output_field] = self.const * np.ones(
            shape=(1, length), dtype=self.dtype
        )
        return data


class AddTimeFeatures(MapTransformation):
    """
    Adds a set of time features.

    If `is_train=True` the feature matrix has the same length as the `target`
    field. If `is_train=False` the feature matrix has length
    `len(target) + pred_length`

    Parameters
    ----------
    start_field
        Field with the start time stamp of the time series
    target_field
        Field with the array containing the time series values
    output_field
        Field name for result.
    time_features
        list of time features to use.
    pred_length
        Prediction length
    """

    @validated()
    def __init__(
        self,
        start_field: str,
        target_field: str,
        output_field: str,
        time_features: List[TimeFeature],
        pred_length: int,
        dtype: Type = np.float32,
    ) -> None:
        self.date_features = time_features
        self.pred_length = pred_length
        self.start_field = start_field
        self.target_field = target_field
        self.output_field = output_field
        self.dtype = dtype

    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        if not self.date_features:
            data[self.output_field] = None
            return data

        start = data[self.start_field]
        length = target_transformation_length(
            data[self.target_field], self.pred_length, is_train=is_train
        )

        index = pd.period_range(start, periods=length, freq=start.freq)

        data[self.output_field] = np.vstack(
            [feat(index) for feat in self.date_features]
        ).astype(self.dtype)

        return data


class AddAgeFeature(MapTransformation):
    """
    Adds an 'age' feature to the data_entry.

    The age feature starts with a small value at the start of the time series
    and grows over time.

    If `is_train=True` the age feature has the same length as the `target`
    field.
    If `is_train=False` the age feature has length len(target) + pred_length

    Parameters
    ----------
    target_field
        Field with target values (array) of time series
    output_field
        Field name to use for the output.
    pred_length
        Prediction length
    log_scale
        If set to true the age feature grows logarithmically otherwise linearly
        over time.
    """

    @validated()
    def __init__(
        self,
        target_field: str,
        output_field: str,
        pred_length: int,
        log_scale: bool = True,
        dtype: Type = np.float32,
    ) -> None:
        self.pred_length = pred_length
        self.target_field = target_field
        self.feature_name = output_field
        self.log_scale = log_scale
        self._age_feature = np.zeros(0)
        self.dtype = dtype

    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        length = target_transformation_length(
            data[self.target_field], self.pred_length, is_train=is_train
        )

        if self.log_scale:
            age = np.log10(2.0 + np.arange(length, dtype=self.dtype))
        else:
            age = np.arange(length, dtype=self.dtype)

        data[self.feature_name] = age.reshape((1, length))

        return data


class AddAggregateLags(MapTransformation):
    """
    Adds aggregate lags as a feature to the data_entry.

    Aggregates the original time series to a new frequency and selects
    the aggregated lags of interest. It does not use aggregate lags that
    need the last `prediction_length` values to be computed. Therefore
    the transformation is applicable to both training and inference.

    If `is_train=True` the lags have the same length as the `target` field.
    If `is_train=False` the lags have length len(target) + pred_length

    Parameters
    ----------
    target_field
        Field with target values (array) of time series
    output_field
        Field name to use for the output.
    pred_length
        Prediction length.
    base_freq
        Base frequency, i.e., the frequency of the original time series.
    agg_freq
        Aggregate frequency, i.e., the frequency of the aggregate time series.
    agg_lags
        List of aggregate lags given in the aggregate frequency. If some of them
        are invalid (need some of the last `prediction_length` values to be
        computed) they are ignored.
    agg_fun
        Aggregation function. Default is 'mean'.
    """

    @validated()
    def __init__(
        self,
        target_field: str,
        output_field: str,
        pred_length: int,
        base_freq: str,
        agg_freq: str,
        agg_lags: List[int],
        agg_fun: str = "mean",
        dtype: Type = np.float32,
    ) -> None:
        self.pred_length = pred_length
        self.target_field = target_field
        self.feature_name = output_field
        self.base_freq = base_freq
        self.agg_freq = agg_freq
        self.agg_lags = agg_lags
        self.agg_fun = agg_fun
        self.dtype = dtype

        self.ratio = pd.Timedelta(self.agg_freq) / pd.Timedelta(self.base_freq)
        assert self.ratio.is_integer() and self.ratio >= 1, (
            "The aggregate frequency should be a multiple of the base"
            " frequency."
        )
        self.ratio = int(self.ratio)

        self.half_window = (self.ratio - 1) // 2
        self.valid_lags = [
            x
            for x in self.agg_lags
            if x > (self.pred_length - 1 + self.half_window) / self.ratio
        ]

        if set(self.agg_lags) - set(self.valid_lags):
            print(
                "The aggregate lags"
                f" {set(self.agg_lags) - set(self.valid_lags)} of frequency"
                f" {self.agg_freq} are ignored."
            )

    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        assert self.base_freq == data["start"].freq

        # convert to pandas Series for easier indexing and aggregation
        if is_train:
            t = data[self.target_field]
        else:
            t = np.concatenate(
                [data[self.target_field], np.zeros(shape=(self.pred_length,))],
                axis=0,
            )

        # convert to pandas Series for easier rolling window aggregation
        t_agg = (pd.Series(t).rolling(self.ratio).agg(self.agg_fun))[
            self.ratio - 1 :
        ]

        # compute the aggregate lags for each time point of the time series
        agg_vals = np.concatenate(
            [
                np.zeros(
                    (max(self.valid_lags) * self.ratio + self.half_window + 1,)
                ),
                t_agg.values,
            ],
            axis=0,
        )
        lags = np.vstack(
            [
                agg_vals[
                    -(l * self.ratio - self.half_window + len(t)) : -(
                        l * self.ratio - self.half_window
                    )
                    if -(l * self.ratio - self.half_window) != 0
                    else None
                ]
                for l in self.valid_lags
            ]
        )

        # update the data entry
        data[self.feature_name] = np.nan_to_num(lags)

        assert data[self.feature_name].shape == (
            len(self.valid_lags),
            len(data[self.target_field]) + self.pred_length * (not is_train),
        )

        return data


class CountTrailingZeros(SimpleTransformation):
    """
    Add the number of 'trailing' zeros in each univariate time series as a
    feature, to be used when dealing with sparse (intermittent) time series.

    For example, for 1-d a time series `[0, 0, 2, 3, 0]`, the number of
    trailing zeros will be 1. If an n-dimensional array is provided, the first
    1-d array along the `axis` dimension will be checked for trailing zeros.
    For example, if axis is set to 1 for a 3-d array A, the transformation
    will return the number of trailing zeros in `A[0, :, 0]`.

    Parameters
    ----------
    new_field
        Name of the new field to be created, which will contain the number of
        trailing zeros.
    target_field
        Field with target values (array) of time series
    as_array
        if True, the returned field will be a numpy array of shape (1,)
    """

    @validated()
    def __init__(
        self,
        new_field: str = "trailing_zeros",
        target_field: str = FieldName.TARGET,
        axis: int = -1,
        as_array: bool = False,
    ) -> None:
        self.target_field = target_field
        self.new_field = new_field
        self.axis = axis
        self.as_array = as_array

    def transform(self, data: DataEntry) -> DataEntry:
        target = data[self.target_field]
        if len(target) == 0:
            return data
        while target.ndim > 1:
            slice_: List[Union[int, slice]] = [0] * target.ndim
            slice_[self.axis] = slice(None)
            target = target[slice_]
        trailing_zeros = len(target) - len(np.trim_zeros(target, "b"))
        data[self.new_field] = (
            np.array([trailing_zeros]) if self.as_array else trailing_zeros
        )
        return data



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import NamedTuple, Optional, Iterable, Dict, Any
import logging

import numpy as np
import pytorch_lightning as pl
import torch.nn as nn

from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.env import env
from gluonts.itertools import Cached
from gluonts.model import Estimator, Predictor
from gluonts.torch.model.predictor import PyTorchPredictor
from gluonts.transform import Transformation

logger = logging.getLogger(__name__)


class TrainOutput(NamedTuple):
    transformation: Transformation
    trained_net: nn.Module
    trainer: pl.Trainer
    predictor: PyTorchPredictor


class PyTorchLightningEstimator(Estimator):
    """
    An `Estimator` type with utilities for creating PyTorch-Lightning-based
    models.

    To extend this class, one needs to implement three methods:
    `create_transformation`, `create_training_network`, `create_predictor`,
    `create_training_data_loader`, and `create_validation_data_loader`.
    """

    @validated()
    def __init__(
        self,
        trainer_kwargs: Dict[str, Any],
        lead_time: int = 0,
    ) -> None:
        super().__init__(lead_time=lead_time)
        self.trainer_kwargs = trainer_kwargs

    def create_transformation(self) -> Transformation:
        """
        Create and return the transformation needed for training and inference.

        Returns
        -------
        Transformation
            The transformation that will be applied entry-wise to datasets,
            at training and inference time.
        """
        raise NotImplementedError

    def create_lightning_module(self) -> pl.LightningModule:
        """
        Create and return the network used for training (i.e., computing the
        loss).

        Returns
        -------
        pl.LightningModule
            The network that computes the loss given input data.
        """
        raise NotImplementedError

    def create_predictor(
        self,
        transformation: Transformation,
        module,
    ) -> PyTorchPredictor:
        """
        Create and return a predictor object.

        Parameters
        ----------
        transformation
            Transformation to be applied to data before it goes into the model.
        module
            A trained `pl.LightningModule` object.

        Returns
        -------
        Predictor
            A predictor wrapping a `nn.Module` used for inference.
        """
        raise NotImplementedError

    def create_training_data_loader(
        self, data: Dataset, module, **kwargs
    ) -> Iterable:
        """
        Create a data loader for training purposes.

        Parameters
        ----------
        data
            Dataset from which to create the data loader.
        module
            The `pl.LightningModule` object that will receive the batches from
            the data loader.

        Returns
        -------
        Iterable
            The data loader, i.e. and iterable over batches of data.
        """
        raise NotImplementedError

    def create_validation_data_loader(
        self, data: Dataset, module, **kwargs
    ) -> Iterable:
        """
        Create a data loader for validation purposes.

        Parameters
        ----------
        data
            Dataset from which to create the data loader.
        module
            The `pl.LightningModule` object that will receive the batches from
            the data loader.

        Returns
        -------
        Iterable
            The data loader, i.e. and iterable over batches of data.
        """
        raise NotImplementedError

    def train_model(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        from_predictor: Optional[PyTorchPredictor] = None,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        ckpt_path: Optional[str] = None,
        **kwargs,
    ) -> TrainOutput:
        transformation = self.create_transformation()

        with env._let(max_idle_transforms=max(len(training_data), 100)):
            transformed_training_data = transformation.apply(
                training_data, is_train=True
            )
            if cache_data:
                transformed_training_data = Cached(transformed_training_data)

            training_network = self.create_lightning_module()

            training_data_loader = self.create_training_data_loader(
                transformed_training_data,
                training_network,
                shuffle_buffer_length=shuffle_buffer_length,
            )

        validation_data_loader = None

        if validation_data is not None:
            with env._let(max_idle_transforms=max(len(validation_data), 100)):
                transformed_validation_data = transformation.apply(
                    validation_data, is_train=True
                )
                if cache_data:
                    transformed_validation_data = Cached(
                        transformed_validation_data
                    )

                validation_data_loader = self.create_validation_data_loader(
                    transformed_validation_data,
                    training_network,
                )

        if from_predictor is not None:
            training_network.load_state_dict(
                from_predictor.network.state_dict()
            )

        monitor = "train_loss" if validation_data is None else "val_loss"
        checkpoint = pl.callbacks.ModelCheckpoint(
            monitor=monitor, mode="min", verbose=True
        )

        custom_callbacks = self.trainer_kwargs.get("callbacks", [])
        callbacks = [checkpoint] + custom_callbacks
        trainer_kwargs = {**self.trainer_kwargs, "callbacks": callbacks}
        trainer = pl.Trainer(**trainer_kwargs)

        trainer.fit(
            model=training_network,
            train_dataloaders=training_data_loader,
            val_dataloaders=validation_data_loader,
            ckpt_path=ckpt_path,
        )

        logger.info(f"Loading best model from {checkpoint.best_model_path}")
        best_model = training_network.load_from_checkpoint(
            checkpoint.best_model_path
        )

        return TrainOutput(
            transformation=transformation,
            trained_net=best_model,
            trainer=trainer,
            predictor=self.create_predictor(transformation, best_model),
        )

    @staticmethod
    def _worker_init_fn(worker_id):
        np.random.seed(np.random.get_state()[1][0] + worker_id)

    def train(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        ckpt_path: Optional[str] = None,
        **kwargs,
    ) -> PyTorchPredictor:
        return self.train_model(
            training_data,
            validation_data,
            shuffle_buffer_length=shuffle_buffer_length,
            cache_data=cache_data,
            ckpt_path=ckpt_path,
        ).predictor

    def train_from(
        self,
        predictor: Predictor,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        ckpt_path: Optional[str] = None,
    ) -> PyTorchPredictor:
        assert isinstance(predictor, PyTorchPredictor)
        return self.train_model(
            training_data,
            validation_data,
            from_predictor=predictor,
            shuffle_buffer_length=shuffle_buffer_length,
            cache_data=cache_data,
            ckpt_path=ckpt_path,
        ).predictor



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import mxnet as mx
import numpy as np
from mxnet import nd

from gluonts.core.component import validated
from gluonts.mx.model.tpp import distribution
from gluonts.mx.model.tpp.distribution.base import TPPDistributionOutput
from gluonts.mx import Tensor
from gluonts.mx.distribution import CategoricalOutput


class DeepTPPNetworkBase(mx.gluon.HybridBlock):
    """
    Temporal point process model based on a recurrent neural network.

    Parameters
    ----------
    num_marks
        Number of discrete marks (correlated processes), that are available
        in the data.
    interval_length
        The length of the total time interval that is in the prediction
        range. Note that in contrast to discrete-time models in the rest
        of GluonTS, the network is trained to predict an interval, in
        continuous time.
    time_distr_output
        Output distribution for the inter-arrival times. Available
        distributions can be found in gluonts.model.tpp.distribution.
    embedding_dim
        Dimension of vector embeddings of marks (used only as input).
    num_hidden_dimensions
        Number of hidden units in the RNN.
    output_scale
        Positive scaling applied to the inter-event times. You should provide
        this argument if the average inter-arrival time is much larger than 1.
    apply_log_to_rnn_inputs
        Apply logarithm to inter-event times that are fed into the RNN.
    """

    @validated()
    def __init__(
        self,
        num_marks: int,
        interval_length: float,
        time_distr_output: TPPDistributionOutput = distribution.WeibullOutput(),  # noqa: E501
        embedding_dim: int = 5,
        num_hidden_dimensions: int = 10,
        output_scale: Optional[Tensor] = None,
        apply_log_to_rnn_inputs: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.num_marks = num_marks
        self.interval_length = interval_length
        self.rnn_hidden_size = num_hidden_dimensions
        self.output_scale = output_scale
        self.apply_log_to_rnn_inputs = apply_log_to_rnn_inputs

        with self.name_scope():
            self.embedding = mx.gluon.nn.Embedding(
                input_dim=num_marks, output_dim=embedding_dim
            )
            self.rnn = mx.gluon.rnn.GRU(
                num_hidden_dimensions,
                input_size=embedding_dim + 1,
                layout="NTC",
            )
            # Conditional distribution over the inter-arrival times
            self.time_distr_output = time_distr_output
            self.time_distr_args_proj = self.time_distr_output.get_args_proj()
            # Conditional distribution over the marks
            if num_marks > 1:
                self.mark_distr_output = CategoricalOutput(num_marks)
                self.mark_distr_args_proj = (
                    self.mark_distr_output.get_args_proj()
                )

    def hybridize(self, active=True, **kwargs):
        if active:
            raise NotImplementedError(
                "DeepTPP blocks do not support hybridization"
            )


class DeepTPPTrainingNetwork(DeepTPPNetworkBase):
    def hybrid_forward(
        self,
        F,
        target: Tensor,
        valid_length: Tensor,
        **kwargs,
    ) -> Tensor:
        """
        Computes the negative log likelihood loss for the given sequences.

        As the model is trained on past (resp. future) or context
        (resp. prediction) "intervals" as opposed to fixed-length "sequences",
        the number of data points available varies across observations. To
        account for this, data is made available to the training network as a
        "ragged" tensor. The number of valid entries in each sequence is
        provided in a separate variable, :code:`xxx_valid_length`.

        Parameters
        ----------
        F
            MXNet backend.
        target
            Tensor with observations.
            Shape: (batch_size, past_max_sequence_length, target_dim).
        valid_length
            The `valid_length` or number of valid entries in the past_target
            Tensor. Shape: (batch_size,)

        Returns
        -------
        Tensor
            Loss tensor. Shape: (batch_size,).
        """
        if F is mx.sym:
            raise ValueError(
                "The DeepTPP model currently doesn't support hybridization."
            )

        batch_size = target.shape[0]
        # IMPORTANT: We add an additional zero at the end of each sequence
        # It will be used to store the time until the end of the interval
        target = F.concat(target, F.zeros((batch_size, 1, 2)), dim=1)
        # (N, T + 1, 2)

        ia_times, marks = F.split(
            target, num_outputs=2, axis=-1
        )  # inter-arrival times, marks
        marks = marks.squeeze(axis=-1)  # (N, T + 1)

        valid_length = valid_length.reshape(-1).astype(
            ia_times.dtype
        )  # make sure shape is (batch_size,)

        if self.apply_log_to_rnn_inputs:
            ia_times_input = ia_times.clip(1e-8, np.inf).log()
        else:
            ia_times_input = ia_times
        rnn_input = F.concat(ia_times_input, self.embedding(marks), dim=-1)
        rnn_output = self.rnn(rnn_input)  # (N, T + 1, H)

        rnn_init_state = F.zeros([batch_size, 1, self.rnn_hidden_size])
        history_emb = F.slice_axis(
            F.concat(rnn_init_state, rnn_output, dim=1),
            axis=1,
            begin=0,
            end=-1,
        )  # (N, T + 1, H)

        # Augment ia_times by adding the time remaining until interval_length
        # Afterwards, each row of ia_times will sum up to interval_length
        ia_times = ia_times.squeeze(axis=-1)  # (N, T + 1)
        time_remaining = self.interval_length - ia_times.sum(-1)  # (N)
        # Equivalent to ia_times[F.arange(N), valid_length] = time_remaining
        indices = F.stack(F.arange(batch_size), valid_length)
        time_remaining_tensor = F.scatter_nd(
            time_remaining, indices, ia_times.shape
        )
        ia_times_aug = ia_times + time_remaining_tensor

        time_distr_args = self.time_distr_args_proj(history_emb)
        time_distr = self.time_distr_output.distribution(
            time_distr_args, scale=self.output_scale
        )
        log_intensity = time_distr.log_intensity(ia_times_aug)  # (N, T + 1)
        log_survival = time_distr.log_survival(ia_times_aug)  # (N, T + 1)

        if self.num_marks > 1:
            mark_distr_args = self.mark_distr_args_proj(history_emb)
            mark_distr = self.mark_distr_output.distribution(mark_distr_args)
            log_intensity = log_intensity + mark_distr.log_prob(marks)

        def _mask(x, sequence_length):
            return F.SequenceMask(
                data=x,
                sequence_length=sequence_length,
                axis=1,
                use_sequence_length=True,
            )

        log_likelihood = F.sum(
            (
                _mask(log_intensity, valid_length)
                + _mask(log_survival, valid_length + 1)
            ),
            axis=-1,
        )  # (N)

        return -log_likelihood


class DeepTPPPredictionNetwork(DeepTPPNetworkBase):
    @validated()
    def __init__(
        self,
        prediction_interval_length: float,
        num_parallel_samples: int = 100,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.num_parallel_samples = num_parallel_samples
        self.prediction_interval_length = prediction_interval_length

    def hybrid_forward(
        self,
        F,
        past_target: Tensor,
        past_valid_length: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        """
        Draw forward samples from the model. At each step, we sample an inter-
        event time and feed it into the RNN to obtain the parameters for the
        next distribution over the inter-event time.

        Parameters
        ----------
        F
            MXNet backend.
        past_target
            Tensor with past observations.
            Shape: (batch_size, context_length, target_dim). Has to comply
            with :code:`self.context_interval_length`.
        past_valid_length
            The `valid_length` or number of valid entries in the past_target
            Tensor. Shape: (batch_size,)

        Returns
        -------
        sampled_target: Tensor
            Predicted inter-event times and marks.
            Shape: (samples, batch_size, max_prediction_length, target_dim).
        sampled_valid_length: Tensor
            The number of valid entries in the time axis of each sample.
            Shape (samples, batch_size)
        """
        # Variable-length generation (while t < t_max) is a potential problem
        if F is mx.sym:
            raise ValueError(
                "The DeepTPP model currently doesn't support hybridization."
            )

        assert (
            past_target.shape[-1] == 2
        ), "TPP data should have two target_dim, interarrival times and marks"

        batch_size = past_target.shape[0]

        # condition the prediction network on the past events
        past_ia_times, past_marks = F.split(
            past_target, num_outputs=2, axis=-1
        )
        past_valid_length = past_valid_length.reshape(-1).astype(
            past_ia_times.dtype
        )

        if self.apply_log_to_rnn_inputs:
            past_ia_times_input = past_ia_times.clip(1e-8, np.inf).log()
        else:
            past_ia_times_input = past_ia_times
        rnn_input = F.concat(
            past_ia_times_input,
            self.embedding(past_marks.squeeze(axis=-1)),
            dim=-1,
        )
        rnn_output = self.rnn(rnn_input)  # (N, T, H)
        rnn_init_state = F.zeros([batch_size, 1, self.rnn_hidden_size])

        past_history_emb = F.concat(
            rnn_init_state, rnn_output, dim=1
        )  # (N, T + 1, H)

        # Select the history embedding after the last event in the past
        indices = F.stack(F.arange(batch_size), past_valid_length)
        history_emb = F.gather_nd(past_history_emb, indices)  # (N, H)

        num_total_samples = self.num_parallel_samples * batch_size
        history_emb = history_emb.expand_dims(0).repeat(
            self.num_parallel_samples, axis=0
        )  # (S, N, H)
        history_emb = history_emb.reshape(
            [num_total_samples, self.rnn_hidden_size]
        )  # (S * N, H)

        sampled_ia_times_list: List[nd.NDArray] = []
        sampled_marks_list: List[nd.NDArray] = []
        arrival_times = F.zeros([num_total_samples])

        # Time from the last observed event until the past interval end
        past_time_elapsed = past_ia_times.squeeze(axis=-1).sum(-1)
        past_time_remaining = self.interval_length - past_time_elapsed  # (N)
        past_time_remaining_repeat = (
            past_time_remaining.expand_dims(0)
            .repeat(self.num_parallel_samples, axis=0)
            .reshape([num_total_samples])
        )  # (S * N)

        first_step = True
        while F.sum(arrival_times < self.prediction_interval_length) > 0:
            # Sample the next inter-arrival time
            time_distr_args = self.time_distr_args_proj(history_emb)
            time_distr = self.time_distr_output.distribution(
                time_distr_args,
                scale=self.output_scale,
            )
            if first_step:
                # Time from the last event until the next event
                next_ia_times = time_distr.sample(
                    lower_bound=past_time_remaining_repeat
                )
                # Time from the prediction interval start until the next event
                clipped_ia_times = next_ia_times - past_time_remaining_repeat
                sampled_ia_times_list.append(clipped_ia_times)
                arrival_times = arrival_times + clipped_ia_times
                first_step = False
            else:
                next_ia_times = time_distr.sample()
                sampled_ia_times_list.append(next_ia_times)
                arrival_times = arrival_times + next_ia_times

            # Sample the next marks
            if self.num_marks > 1:
                mark_distr_args = self.mark_distr_args_proj(history_emb)
                next_marks = self.mark_distr_output.distribution(
                    mark_distr_args
                ).sample()
            else:
                next_marks = F.zeros([num_total_samples])

            sampled_marks_list.append(next_marks)

            # Pass the generated ia_times & marks into the RNN to obtain
            # the next history embedding
            if self.apply_log_to_rnn_inputs:
                next_ia_times_input = next_ia_times.clip(1e-8, np.inf).log()
            else:
                next_ia_times_input = next_ia_times
            rnn_input = F.concat(
                next_ia_times_input.expand_dims(-1),
                self.embedding(next_marks),
                dim=-1,
            ).expand_dims(1)

            history_emb = self.rnn(rnn_input).squeeze(axis=1)  # (S * N, C)

        sampled_ia_times = F.stack(*sampled_ia_times_list, axis=-1)
        sampled_marks = F.stack(*sampled_marks_list, axis=-1).astype("float32")

        sampled_valid_length = F.sum(
            F.cumsum(sampled_ia_times, axis=1)
            < self.prediction_interval_length,
            axis=-1,
        )

        def _mask(x, sequence_length):
            return F.SequenceMask(
                data=x,
                sequence_length=sequence_length,
                axis=1,
                use_sequence_length=True,
            )

        sampled_ia_times = _mask(sampled_ia_times, sampled_valid_length)
        sampled_marks = _mask(sampled_marks, sampled_valid_length)

        sampled_ia_times = sampled_ia_times.reshape(
            [self.num_parallel_samples, batch_size, -1]
        )
        sampled_marks = sampled_marks.reshape(
            [self.num_parallel_samples, batch_size, -1]
        )
        sampled_valid_length = sampled_valid_length.reshape(
            [self.num_parallel_samples, batch_size]
        )
        sampled_target = F.stack(sampled_ia_times, sampled_marks, axis=-1)
        return sampled_target, sampled_valid_length



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
import tarfile
from pathlib import Path
from typing import Optional

import pytest

from gluonts.util import copy_with, will_extractall_into


def test_copy_with():
    class X:
        def __init__(self, value):
            self.value = value

    a = X(42)
    b = copy_with(a, value=99)

    assert a.value == 42
    assert b.value == 99


@pytest.mark.parametrize(
    "arcname, expect_failure",
    [
        (None, False),
        ("./file.txt", False),
        ("/a/../file.txt", False),
        ("/a/../../file.txt", True),
        ("../file.txt", True),
    ],
)
def test_will_extractall_into(arcname: Optional[str], expect_failure: bool):
    with tempfile.TemporaryDirectory() as tempdir:
        file_path = Path(tempdir) / "a" / "file.txt"
        file_path.parent.mkdir(parents=True)
        file_path.touch()

        with tarfile.open(Path(tempdir) / "archive.tar.gz", "w:gz") as tar:
            tar.add(file_path, arcname=arcname)

        if expect_failure:
            with pytest.raises(PermissionError):
                with tarfile.open(
                    Path(tempdir) / "archive.tar.gz", "r:gz"
                ) as tar:
                    will_extractall_into(tar, Path(tempdir) / "b")
        else:
            with tarfile.open(Path(tempdir) / "archive.tar.gz", "r:gz") as tar:
                will_extractall_into(tar, Path(tempdir) / "b")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import math
from typing import Optional, Tuple

import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.linalg_util import lower_triangular_ones

from .distribution import Distribution, _sample_multiple, getF
from .distribution_output import DistributionOutput


class MultivariateGaussian(Distribution):
    r"""
    Multivariate Gaussian distribution, specified by the mean vector
    and the Cholesky factor of its covariance matrix.

    Parameters
    ----------
    mu
        mean vector, of shape (..., d)
    L
        Lower triangular Cholesky factor of covariance matrix, of shape
        (..., d, d)
    F
        A module that can either refer to the Symbol API or the NDArray
        API in MXNet
    """

    is_reparameterizable = True

    @validated()
    def __init__(self, mu: Tensor, L: Tensor, F=None) -> None:
        self.mu = mu
        self.L = L

    @property
    def F(self):
        return getF(self.mu)

    def __getitem__(self, item):
        raise NotImplementedError()

    @property
    def batch_shape(self) -> Tuple:
        return self.mu.shape[:-1]

    @property
    def event_shape(self) -> Tuple:
        return self.mu.shape[-1:]

    @property
    def event_dim(self) -> int:
        return 1

    def log_prob(self, x: Tensor) -> Tensor:
        # todo add an option to compute loss on diagonal covariance only tosave
        # time
        F = self.F

        # remark we compute d from the tensor but we could ask it to the user
        # alternatively
        d = F.ones_like(self.mu).sum(axis=-1).max()

        residual = (x - self.mu).expand_dims(axis=-1)

        # L^{-1} * (x - mu)
        L_inv_times_residual = F.linalg_trsm(self.L, residual)

        ll = (
            F.broadcast_sub(
                -d / 2 * math.log(2 * math.pi), F.linalg_sumlogdiag(self.L)
            )
            - 1
            / 2
            * F.linalg_syrk(L_inv_times_residual, transpose=True).squeeze()
        )

        return ll

    @property
    def mean(self) -> Tensor:
        return self.mu

    @property
    def variance(self) -> Tensor:
        return self.F.linalg_gemm2(self.L, self.L, transpose_b=True)

    def sample_rep(
        self, num_samples: Optional[int] = None, dtype=np.float32
    ) -> Tensor:
        r"""
        Draw samples from the multivariate Gaussian distributions.
        Internally, Cholesky factorization of the covariance matrix is used:

            sample = L v + mu,

        where L is the Cholesky factor, v is a standard normal sample.

        Parameters
        ----------
        num_samples
            Number of samples to be drawn.
        dtype
            Data-type of the samples.
        Returns
        -------
        Tensor
            Tensor with shape (num_samples, ..., d).
        """

        def s(mu: Tensor, L: Tensor) -> Tensor:
            F = self.F
            samples_std_normal = F.sample_normal(
                mu=F.zeros_like(mu),
                sigma=F.ones_like(mu),
                dtype=dtype,
            ).expand_dims(axis=-1)
            samples = (
                F.linalg_gemm2(L, samples_std_normal).squeeze(axis=-1) + mu
            )
            return samples

        return _sample_multiple(
            s, mu=self.mu, L=self.L, num_samples=num_samples
        )


class MultivariateGaussianOutput(DistributionOutput):
    @validated()
    def __init__(self, dim: int) -> None:
        self.args_dim = {"mu": dim, "Sigma": dim * dim}
        self.distr_cls = MultivariateGaussian
        self.dim = dim
        self.mask = None

    def domain_map(self, F, mu_vector, L_vector):
        # apply softplus to the diagonal of L and mask upper coefficient to
        # make it lower-triangular diagonal matrix whose elements are diagonal
        # elements of L mapped through a softplus
        d = self.dim

        # reshape from vector form (..., d * d) to matrix form(..., d, d)
        L_matrix = L_vector.reshape((-2, d, d, -4), reverse=1)

        L_diag = F.broadcast_mul(
            F.Activation(
                F.broadcast_mul(L_matrix, F.eye(d)), act_type="softrelu"
            ),
            F.eye(d),
        )

        mask = lower_triangular_ones(F, d, offset=1)

        L_low = F.broadcast_mul(L_matrix, mask)

        return mu_vector, L_diag + L_low

    @property
    def event_shape(self) -> Tuple:
        return (self.dim,)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import pytest
import numpy as np

from gluonts.dataset import Dataset
from gluonts.model import Predictor, QuantileForecast
from gluonts.ext.statsforecast import (
    ModelConfig,
    StatsForecastPredictor,
    ADIDAPredictor,
    AutoARIMAPredictor,
    AutoCESPredictor,
    AutoETSPredictor,
    AutoThetaPredictor,
    CrostonClassicPredictor,
    CrostonOptimizedPredictor,
    CrostonSBAPredictor,
    DynamicOptimizedThetaPredictor,
    DynamicThetaPredictor,
    HistoricAveragePredictor,
    HoltPredictor,
    HoltWintersPredictor,
    IMAPAPredictor,
    MSTLPredictor,
    NaivePredictor,
    OptimizedThetaPredictor,
    RandomWalkWithDriftPredictor,
    SeasonalExponentialSmoothingOptimizedPredictor,
    SeasonalExponentialSmoothingPredictor,
    SeasonalNaivePredictor,
    SeasonalWindowAveragePredictor,
    SimpleExponentialSmoothingOptimizedPredictor,
    SimpleExponentialSmoothingPredictor,
    TSBPredictor,
    ThetaPredictor,
    WindowAveragePredictor,
)


@pytest.mark.parametrize(
    "quantile_levels, intervals, forecast_keys, statsforecast_keys",
    [
        (
            [0.5, 0.2, 0.6, 0.9, 0.8, 0.1],
            [0, 20, 60, 80],
            ["mean", "0.5", "0.2", "0.6", "0.9", "0.8", "0.1"],
            ["mean", "lo-0", "lo-60", "hi-20", "hi-80", "hi-60", "lo-80"],
        ),
    ],
)
def test_model_config(
    quantile_levels, intervals, forecast_keys, statsforecast_keys
):
    config = ModelConfig(quantile_levels=quantile_levels)
    assert config.intervals == intervals
    assert config.forecast_keys == forecast_keys
    assert config.statsforecast_keys == statsforecast_keys


@pytest.mark.parametrize(
    "predictor",
    [
        ADIDAPredictor(prediction_length=3),
        AutoARIMAPredictor(
            prediction_length=3, quantile_levels=[0.5, 0.1, 0.9, 0.95]
        ),
        AutoCESPredictor(prediction_length=3, season_length=12),
        AutoETSPredictor(
            prediction_length=3,
            season_length=12,
            quantile_levels=[0.5, 0.1, 0.9, 0.95],
        ),
        AutoThetaPredictor(
            prediction_length=3,
            season_length=12,
            quantile_levels=[0.5, 0.1, 0.9, 0.95],
        ),
        CrostonClassicPredictor(prediction_length=3),
        CrostonOptimizedPredictor(prediction_length=3),
        CrostonSBAPredictor(prediction_length=3),
        DynamicOptimizedThetaPredictor(prediction_length=3),
        DynamicThetaPredictor(prediction_length=3),
        HistoricAveragePredictor(prediction_length=3),
        HoltPredictor(prediction_length=3, season_length=2),
        HoltWintersPredictor(prediction_length=3, season_length=2),
        IMAPAPredictor(prediction_length=3),
        MSTLPredictor(prediction_length=3, season_length=2),
        NaivePredictor(prediction_length=3),
        OptimizedThetaPredictor(prediction_length=3),
        RandomWalkWithDriftPredictor(prediction_length=3),
        SeasonalExponentialSmoothingOptimizedPredictor(
            prediction_length=3, season_length=2
        ),
        SeasonalExponentialSmoothingPredictor(
            prediction_length=3, season_length=2, alpha=0.5
        ),
        SeasonalNaivePredictor(prediction_length=3, season_length=2),
        SeasonalWindowAveragePredictor(
            prediction_length=3, season_length=2, window_size=1
        ),
        SimpleExponentialSmoothingOptimizedPredictor(prediction_length=3),
        SimpleExponentialSmoothingPredictor(prediction_length=3, alpha=0.5),
        TSBPredictor(prediction_length=3, alpha_d=0.5, alpha_p=0.5),
        ThetaPredictor(prediction_length=3),
        WindowAveragePredictor(prediction_length=3, window_size=1),
    ],
)
@pytest.mark.parametrize(
    "dataset",
    [
        [
            dict(
                start=pd.Period("2021-02-03 00", freq="H"),
                target=np.random.normal(loc=10, scale=0.5, size=(100,)),
            )
        ]
    ],
)
def test_predictor_working(
    predictor: StatsForecastPredictor, dataset: Dataset
):
    with TemporaryDirectory() as temp_dir:
        predictor.serialize(Path(temp_dir))
        predictor_copy = Predictor.deserialize(Path(temp_dir))

    assert predictor_copy == predictor
    assert isinstance(predictor_copy.model, predictor_copy.ModelType)

    for forecast in predictor_copy.predict(dataset):
        assert isinstance(forecast, QuantileForecast)
        assert len(forecast.mean) == predictor_copy.prediction_length
        assert predictor_copy.config.forecast_keys == forecast.forecast_keys



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Optional, Union

import mxnet as mx
import numpy as np
import pandas as pd
from gluonts.core.component import validated
from gluonts.model.forecast import (
    Forecast,
    Quantile,
    QuantileForecast,
    SampleForecast,
)
from gluonts.mx.distribution import Distribution


class DistributionForecast(Forecast):
    """
    A `Forecast` object that uses a GluonTS distribution directly. This can for
    instance be used to represent marginal probability distributions for each
    time point -- although joint distributions are also possible, e.g. when
    using MultiVariateGaussian).

    Parameters
    ----------
    distribution
        Distribution object. This should represent the entire prediction
        length, i.e., if we draw `num_samples` samples from the distribution,
        the sample shape should be

            samples = trans_dist.sample(num_samples)
            samples.shape -> (num_samples, prediction_length)

    start_date
        start of the forecast
    info
        additional information that the forecaster may provide e.g. estimated
        parameters, number of iterations ran etc.
    """

    @validated()
    def __init__(
        self,
        distribution: Distribution,
        start_date: pd.Period,
        item_id: Optional[str] = None,
        info: Optional[Dict] = None,
    ) -> None:
        self.distribution = distribution
        self.shape = (
            self.distribution.batch_shape + self.distribution.event_shape
        )
        self.prediction_length = self.shape[0]
        self.item_id = item_id
        self.info = info

        assert isinstance(
            start_date, pd.Period
        ), "start_date should be a pandas Period object"
        self.start_date = start_date

        self._mean = None

    @property
    def mean(self) -> np.ndarray:
        """
        Forecast mean.
        """
        if self._mean is not None:
            return self._mean
        else:
            self._mean = self.distribution.mean.asnumpy()
            return self._mean

    @property
    def mean_ts(self) -> pd.Series:
        """
        Forecast mean, as a pandas.Series object.
        """
        return pd.Series(self.mean, index=self.index)

    def quantile(self, level: Union[float, str]) -> np.ndarray:
        level = Quantile.parse(level).value
        q = self.distribution.quantile(mx.nd.array([level])).asnumpy()[0]
        return q

    def to_sample_forecast(self, num_samples: int = 200) -> SampleForecast:
        return SampleForecast(
            samples=self.distribution.sample(num_samples),
            start_date=self.start_date,
            item_id=self.item_id,
            info=self.info,
        )

    def to_quantile_forecast(self, quantiles: List[Union[float, str]]):
        return QuantileForecast(
            forecast_arrays=np.array([self.quantile(q) for q in quantiles]),
            forecast_keys=quantiles,
            start_date=self.start_date,
            item_id=self.item_id,
            info=self.info,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Any, List

import pytest

from gluonts.mx.trainer import Trainer


def test_epochs() -> None:
    assert_valid_param(
        param_name="epochs", param_values=[0, 1, 42, 1000, 1000]
    )
    assert_invalid_param(
        param_name="epochs",
        param_values=[-2, -1],
        exp_msg="The value of `epochs` should be > 0 (type=value_error)",
    )


@pytest.mark.skip()  # TODO move to callback test
def test_patience() -> None:
    assert_valid_param(param_name="patience", param_values=[0, 1, 10, 100])
    assert_invalid_param(
        param_name="patience",
        param_values=[-2, -1],
        exp_msg="The value of `patience` should be >= 0 (type=value_error)",
    )


def test_learning_rate() -> None:
    assert_valid_param(
        param_name="learning_rate", param_values=[0.42, 17.8, 10.0]
    )
    assert_invalid_param(
        param_name="learning_rate",
        param_values=[-2, -1e-10, 0, float("inf"), float("nan")],
        exp_msg="The value of `learning_rate` should be > 0 (type=value_error)",
    )


@pytest.mark.skip()  # TODO move to callback test
def test_learning_rate_decay_factor() -> None:
    assert_valid_param(
        param_name="learning_rate_decay_factor",
        param_values=[0.1, 1e-10, 0.5, 1 - 1e-10],
    )
    assert_invalid_param(
        param_name="learning_rate_decay_factor",
        param_values=[-2, -1e-10, +1, +5, float("inf"), float("nan"), 0],
        exp_msg="The value of `learning_rate_decay_factor` should be in the (0, 1) range (type=value_error)",
    )


def assert_valid_param(param_name: str, param_values: List[Any]) -> None:
    try:
        for x in param_values:
            Trainer(**{param_name: x})
    except Exception as e:
        pytest.fail(f'Unexpected exception when initializing Trainer: "{e}"')
        raise e


def assert_invalid_param(
    param_name: str, param_values: List[Any], exp_msg: str
) -> None:
    for x in param_values:
        with pytest.raises(AssertionError) as excinfo:
            Trainer(**{param_name: x})
            assert exp_msg in str(excinfo.value)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations

import dataclasses
import itertools
from operator import itemgetter
from typing import Optional, List

import numpy as np
from toolz import first, keymap, valmap, dissoc, merge

from gluonts import maybe
from gluonts.itertools import pluck_attr, rows_to_columns

from ._base import Pad
from ._period import Periods, period
from ._repr import html_table
from ._util import _replace


@dataclasses.dataclass
class SplitFrame:
    _past: dict
    _future: dict
    index: Optional[Periods]
    static: dict
    past_length: int
    future_length: int
    tdims: dict
    metadata: Optional[dict] = None
    default_tdim: int = -1
    _pad: Pad = Pad()

    def __post_init__(self):
        for column in itertools.chain(self._past, self._future):
            self.tdims.setdefault(column, self.default_tdim)

        # this triggers checks for past_length and future_length
        _, _ = self.past, self.future

    @property
    def past(self):
        return TimeFrame(
            self._past,
            index=maybe.map(
                self.index, lambda index: index[: self.past_length]
            ),
            static=self.static,
            length=self.past_length,
            tdims=self.tdims,
            metadata=self.metadata,
            _pad=Pad(self._pad.left, 0),
        )

    @property
    def future(self):
        return TimeFrame(
            self._future,
            index=maybe.map(
                self.index, lambda index: index[self.past_length :]
            ),
            static=self.static,
            length=self.future_length,
            tdims=self.tdims,
            metadata=self.metadata,
            _pad=Pad(0, self._pad.right),
        )

    def __len__(self):
        return self.past_length + self.future_length

    def set(self, name, value, tdim=None):
        tdim = maybe.unwrap_or(tdim, self.default_tdim)
        assert value.shape[tdim] == len(self)

        past, future = np.split(
            value,
            [self.past_length],
            axis=tdim,
        )

        return _replace(
            past=merge(self.past, {name: past}),
            future=merge(self.future, {name: future}),
            tdims=merge(self.tdims, {name: tdim}),
        )

    def set_like(self, ref: str, column, value, tdim=None):
        is_past = ref in self._past
        is_future = ref in self._future

        if is_past:
            if is_future:
                return self.set(column, value, tdim)
            else:
                return self.set_past(column, value, tdim)
        elif is_future:
            return self.set_future(column, value, tdim)

        raise KeyError(f"Ref {ref} is neither past nor future")

    def set_past(self, name, value, tdim=None):
        tdim = maybe.unwrap_or(tdim, self.default_tdim)
        assert value.shape[tdim] == self.past_length
        assert self.tdims.get(name, tdim) == tdim

        return _replace(
            past=merge(self.past, {name: value}),
            tdims=merge(self.tdims, {name: tdim}),
        )

    def set_future(self, name, value, tdim=None):
        tdim = maybe.unwrap_or(tdim, self.default_tdim)
        assert value.shape[tdim] == self.future_length
        assert self.tdims.get(name, tdim) == tdim

        return _replace(
            future=merge(self.future, {name: value}),
            tdims=merge(self.tdims, {name: tdim}),
        )

    def remove(self, column):
        return _replace(
            self,
            columns=dissoc(self.columns, column),
            tdims=dissoc(self.tdims, column),
        )

    def _repr_html_(self):
        past = self.past._table_columns()
        future = self.future._table_columns()

        length = max(
            len(first(past.values())) if past else 0,
            len(first(future.values())) if future else 0,
        )

        def pad(col):
            to_pad = length - len(col)

            return list(col) + [""] * to_pad

        past = valmap(pad, past)
        future = valmap(pad, future)

        past = keymap(lambda key: f"past_{key}" if key else "past", past)
        future = keymap(
            lambda key: f"future_{key}" if key else "future", future
        )

        return html_table({**past, "|": ["|"] * length, **future})

    def as_dict(self):
        past = keymap(lambda key: f"past_{key}", self._past)
        future = keymap(lambda key: f"future_{key}", self._future)

        return {**past, **future, **self.static}

    def resize(
        self,
        past_length: Optional[int] = None,
        future_length: Optional[int] = None,
        pad_value=0.0,
    ) -> SplitFrame:
        index = self.index

        past_length = maybe.unwrap_or(past_length, self.past_length)
        future_length = maybe.unwrap_or(future_length, self.future_length)

        if index is not None:
            start = index[0] + (self.past_length + past_length)
            index = start.periods(past_length + future_length)

        return _replace(
            self,
            _past=self.past.resize(
                past_length, pad_value, pad="l", skip="l"
            ).columns,
            past_length=maybe.unwrap_or(past_length, self.past_length),
            _future=self.future.resize(
                future_length, pad_value, pad="r", skip="r"
            ).columns,
            future_length=maybe.unwrap_or(future_length, self.future_length),
            index=index,
        )

    def with_index(self, index):
        return _replace(self, index=index)

    @staticmethod
    def _batch(split_frames: List[SplitFrame]) -> BatchSplitFrame:
        ref = split_frames[0]
        pluck = pluck_attr(split_frames)

        return BatchSplitFrame(
            _past=rows_to_columns(pluck("_past"), np.stack),  # type: ignore
            _future=rows_to_columns(pluck("_future"), np.stack),  # type: ignore
            index=pluck("index"),
            static=rows_to_columns(pluck("static"), np.stack),  # type: ignore
            past_length=ref.past_length,
            future_length=ref.future_length,
            tdims=ref.tdims,
            metadata=pluck("metadata"),
            _pad=pluck("_pad"),
        )


@dataclasses.dataclass
class BatchSplitFrame:
    _past: dict
    _future: dict
    index: List[Optional[Periods]]
    static: dict
    past_length: int
    future_length: int
    tdims: dict
    metadata: List[Optional[dict]]
    _pad: List[Pad]

    @property
    def batch_size(self):
        return len(self.index)

    def __len__(self):
        return self.past_length + self.future_length

    @property
    def past(self):
        return BatchTimeFrame(
            columns=self._past,
            static=self.static,
            length=self.past_length,
            index=[
                maybe.map(index, itemgetter(slice(None, self.past_length)))
                for index in self.index
            ],
            tdims=self.tdims,
            metadata=self.metadata,
            _pad=self._pad,
        )

    @property
    def future(self):
        return BatchTimeFrame(
            columns=self._future,
            static=self.static,
            length=self.future_length,
            index=[
                maybe.map(index, itemgetter(slice(self.past_length, None)))
                for index in self.index
            ],
            tdims=self.tdims,
            metadata=self.metadata,
            _pad=self._pad,
        )

    def items(self):
        return BatchSplitFrameItems(self)

    def as_dict(self):
        return SplitFrame.as_dict(self)


@dataclasses.dataclass(repr=False)
class BatchSplitFrameItems:
    data: BatchSplitFrame

    def __len__(self):
        return self.data.batch_size

    def __getitem__(self, idx):
        tdims = self.data.tdims

        if isinstance(idx, int):
            cls = SplitFrame
            tdims = valmap(lambda tdim: tdim - 1 if tdim >= 0 else tdim, tdims)
        else:
            cls = BatchSplitFrame

        return cls(
            _past=valmap(itemgetter(idx), self.data._past),
            _future=valmap(itemgetter(idx), self.data._future),
            index=self.data.index[idx],
            static=valmap(itemgetter(idx), self.data.static),
            tdims=tdims,
            metadata=self.data.metadata[idx],
            _pad=self.data._pad[idx],
            past_length=self.data.past_length,
            future_length=self.data.future_length,
        )


def split_frame(
    full=None,
    *,
    past=None,
    future=None,
    past_length=None,
    future_length=None,
    static=None,
    index=None,
    start=None,
    freq=None,
    tdims=None,
    metadata=None,
    default_tdim=-1,
):
    """Create a ``zebras.SplitFrame`` where columns can either be `past`,
    `future` or `full`, which spans both past and future.

    ``past_length`` and ``future_length`` is derived from the input data if
    possible or default to ``0`` in case no respective data is available. It is
    possible to set these values explicitly for enforcing consistency or to
    provide a length even though no time series spans that range.

    Parameters
    ----------
    full, optional
        Time series columns that span past and future.
    past, optional
        Time series columns that are past only.
    future, optional
        Time series columns that are future only.
    past_length, optional
        Set length of the past section, derived from data if not provided.
    future_length, optional
        Set length of the future section, derived from data if not provided.
    static, optional
        Values that are independent of time.
    index, optional
        A ``zebras.Periods`` object representing timestamps.
        Must have the same length as full range.
    start, optional
        The start time represented by a string (e.g., "2023-01-01"),
        or a ``zebras.Period`` object. An index will be constructed using
        this start time and the specified frequency
    freq, optional
        The frequency to use for constructing the index.
    tdims, optional
        A dictionary specifying the time dimension for each column, this
        applies to past, future and full.
    default_tdim, optional
        The default time dimension, by default -1
    metadata, optional
        A dictionary of metadata associated with the TimeFrame, by default None
    Returns
    -------
        A ``zebras.SplitFrame`` object.
    """
    full = valmap(np.array, maybe.unwrap_or_else(full, dict))
    past = valmap(np.array, maybe.unwrap_or_else(past, dict))
    future = valmap(np.array, maybe.unwrap_or_else(future, dict))
    static = valmap(np.array, maybe.unwrap_or_else(static, dict))
    tdims = maybe.unwrap_or_else(tdims, dict)

    # Resolve `past_length` and `future_length` if not directly set from
    # provided data. If no data is passed for either field, the value is still
    # `None` after this.
    if past_length is None and past:
        column = first(past)
        past_length = past[column].shape[tdims.get(column, default_tdim)]

    if future_length is None and future:
        column = first(future)
        future_length = future[column].shape[tdims.get(column, default_tdim)]

    if full:
        column = first(full)
        full_length = full[column].shape[tdims.get(column, default_tdim)]

        if maybe.or_(past_length, future_length) is None:
            raise ValueError(
                "Cannot determine past and future length if only "
                "`full` is provided."
            )
    # No data and no lengths are passed
    elif maybe.or_(past_length, future_length) is None:
        past_length = 0
        future_length = 0
        full_length = 0
    else:
        # If past_length and/or future_length is provided, but no `full` data
        # is given, then we first resolve past and future length and then just
        # calculate full_length later.
        full_length = None

    if past_length is None:
        past_length = maybe.map_or(
            full_length,
            lambda fl: fl - future_length,
            0,
        )
    elif future_length is None:
        future_length = maybe.map_or(
            full_length,
            lambda fl: fl - past_length,
            0,
        )
    full_length = past_length + future_length

    # create copies to not mutate user provided dicts
    past = dict(past)
    future = dict(future)

    for name, data in full.items():
        tdim = tdims.get(name, default_tdim)
        assert data.shape[tdim] == full_length

        past_data, future_data = np.split(data, [past_length], axis=tdim)
        past[name] = past_data
        future[name] = future_data

    sf = SplitFrame(
        _past=past,
        _future=future,
        index=index,
        static=static,
        tdims=tdims,
        past_length=past_length,
        future_length=future_length,
        default_tdim=default_tdim,
        metadata=metadata,
    )

    if sf.index is None and start is not None:
        if freq is not None:
            start = period(start, freq)

        return sf.with_index(start.periods(len(sf)))

    return sf


# We defer these imports to avoid circular imports.
from ._time_series import TimeSeries  # noqa
from ._time_frame import BatchTimeFrame, TimeFrame  # noqa



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx
import pytest

from gluonts.mx.representation import Representation

cases = [
    (
        mx.nd.array(
            [
                [1.0] * 50,
                [0.0] * 25 + [3.0] * 25,
                [2.0] * 49 + [1.5] * 1,
                [0.0] * 50,
                [1.0] * 50,
            ]
        ),
        mx.nd.array(
            [
                [1.0] * 50,
                [0.0] * 25 + [1.0] * 25,
                [0.0] * 49 + [1.0] * 1,
                [1.0] * 50,
                [0.0] * 50,
            ]
        ),
    ),
    (
        mx.nd.array(
            [
                [120.0] * 25 + [150.0] * 25,
                [0.0] * 10 + [3.0] * 20 + [61.0] * 20,
                [0.0] * 50,
                [2e-2] * 10 + [0.0] * 30 + [3e-2] * 10,
            ]
        ),
        mx.nd.array(
            [
                [1.0] * 25 + [1.0] * 25,
                [0.0] * 10 + [1.0] * 20 + [1.0] * 20,
                [0.0] * 50,
                [1.0] * 10 + [0.0] * 30 + [1.0] * 10,
            ]
        ),
    ),
    (
        mx.nd.random.normal(shape=(5, 30)),
        mx.nd.zeros(shape=(5, 30)),
    ),
]


@pytest.mark.parametrize("target, observed", cases)
def test_rep(target, observed):
    s = Representation()
    target_scaled, scale, _ = s(target, observed, None, [])

    assert mx.nd.norm(target - target_scaled) == 0



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.exceptions import GluonTSDataError
from gluonts.mx.model.deepar import DeepAREstimator
from gluonts.mx.trainer import Trainer


@pytest.mark.parametrize("dataset", [[]])
def test_deepar_no_batches(dataset):
    estimator = DeepAREstimator(
        prediction_length=10,
        freq="H",
        trainer=Trainer(epochs=1, num_batches_per_epoch=1),
    )

    with pytest.raises(GluonTSDataError):
        estimator.train(dataset)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Flat encoding for serde.

`flat.encode` always returns a flat dictionary, where keys contain information
for nested objects::

    class Inner(NamedTuple):
        val: int

    class Outer(NamedTuple):
        inner: Inner

    value = Outer(inner=Inner(val=42))

    assert encode(value) == {
        '()': '__main__.Outer',
        'inner.()': '__main__.Inner',
        'inner.val': 42},
    }
"""

from collections import defaultdict
from itertools import count
from typing import Any

from toolz.dicttoolz import keymap, valmap

from ._base import encode as base_encode, Kind, decode as base_decode


def join(a, b, sep="."):
    """
    Joins `a` and `b` using `sep`.
    """
    if not a:
        return b

    return f"{a}{sep}{b}"


def _encode(data: Any, path: str, result: dict):
    if isinstance(data, dict) and "__kind__" in data:
        kind = data["__kind__"]

        if kind == Kind.Instance:
            result[join(path, "()")] = data["class"]

            for n, arg in enumerate(data.get("args", [])):
                _encode(arg, join(path, n), result)

            for name, value in data["kwargs"].items():
                _encode(value, join(path, name), result)

        elif kind == Kind.Stateful:
            result[join(path, "!")] = data["class"]

            for name, value in data["kwargs"].items():
                _encode(value, join(path, name), result)

        elif kind == Kind.Type:
            result[join(path, "#")] = data["class"]

        else:
            raise ValueError(f"Unsupported kind `{kind}`.")

    else:
        result[path] = data


def _asdict(trie):
    if not isinstance(trie, defaultdict):
        return trie
    return {k: _asdict(v) for k, v in trie.items()}


def nest(data):
    Trie = lambda: defaultdict(Trie)  # noqa
    trie = Trie()

    for parts, value in data.items():
        *parts, key = parts

        # walk
        current = trie
        for part in parts:
            current = current[part]

        current[key] = value

    return _asdict(trie)


def get_args(data):
    args = []
    for idx in map(str, count(start=0)):
        if idx not in data:
            return args
        args.append(_translate(data.pop(idx)))


def _translate(data):
    if isinstance(data, dict):
        if "()" in data:
            ty = data.pop("()")
            args = get_args(data)
            kwargs = valmap(_translate, data)

            return {
                "__kind__": "instance",
                "class": ty,
                "args": args,
                "kwargs": kwargs,
            }

        if "!" in data:
            ty = data.pop("!")
            kwargs = valmap(_translate, data)

            return {
                "__kind__": "stateful",
                "class": ty,
                "kwargs": kwargs,
            }

        if "#" in data:
            return {
                "__kind__": "type",
                "class": data.pop("#"),
            }

        return valmap(_translate, data)

    if isinstance(data, list):
        return list(map(_translate, data))

    return data


def decode(data: dict) -> Any:
    def split_path(s):
        return tuple(s.split("."))

    nested = nest(keymap(split_path, data))
    encoded = _translate(nested)
    return base_decode(encoded)


def encode(obj) -> dict:
    """
    Encode a given object into a flat-dictionary.

    It uses the default-encoding, to then flatten the output.
    """
    encoded = base_encode(obj)

    result: dict = {}
    _encode(encoded, "", result)
    return result


def clone(data, kwargs=None):
    """
    Create a copy of a given value, by calling `encode` and `decode` on it.

    If `kwargs` is provided, it's possible to overwrite nested values.
    """
    encoded = encode(data)
    if kwargs:
        encoded.update(kwargs)

    return decode(encoded)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
import re
from collections import defaultdict

from toolz.dicttoolz import valmap

from gluonts.core import serde


def split_by_prefix(data: dict) -> dict:
    rx = re.compile(r"\$(\w+)\.")

    namespace: dict = defaultdict(dict)

    def split_key(key):
        key_parts = rx.split(key, 1)
        if len(key_parts) == 1:
            prefix = ""
            suffix = key_parts[0]
        else:
            _, prefix, suffix = key_parts

        return prefix, suffix

    for key, value in data.items():
        prefix, suffix = split_key(key)
        namespace[prefix][suffix] = value

    return dict(namespace)


def decode_nested_parameters(parameters: dict) -> dict:
    inputs = split_by_prefix(parameters)
    return valmap(serde.flat.decode, inputs)


def encode_nested_parameters(obj) -> dict:
    return serde.flat.encode(obj)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Tuple

from mxnet import gluon

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .deterministic import Deterministic
from .distribution import Distribution, getF, nans_like
from .distribution_output import DistributionOutput
from .mixture import MixtureDistribution


class NanMixture(MixtureDistribution):
    r"""
    A mixture distribution of a NaN-valued Deterministic distribution and
    Distribution

    Parameters
    ----------
    nan_prob
        A tensor of the probabilities of missing values. The entries should all
        be positive and smaller than 1. All axis should either coincide with
        the ones from the component distributions, or be 1 (in which case, the
        NaN probability is shared across the axis).
    distribution
        A Distribution object representing the Distribution of non-NaN values.
        Distributions can be of different types. Each component's support
        should be made of tensors of shape (..., d).
    F
        A module that can either refer to the Symbol API or the NDArray
        API in MXNet
    """

    is_reparameterizable = False

    @validated()
    def __init__(
        self, nan_prob: Tensor, distribution: Distribution, F=None
    ) -> None:
        F = getF(nan_prob)

        mixture_probs = F.stack(1 - nan_prob, nan_prob, axis=-1)
        super().__init__(
            mixture_probs=mixture_probs,
            components=[
                distribution,
                Deterministic(value=nans_like(nan_prob)),
            ],
        )

    @property
    def distribution(self):
        return self.components[0]

    @property
    def nan_prob(self):
        return self.mixture_probs.slice_axis(axis=-1, begin=1, end=2).squeeze(
            axis=-1
        )

    def log_prob(self, x: Tensor) -> Tensor:
        F = self.F

        # masking data NaN's with ones to prevent NaN gradients
        x_non_nan = F.where(x != x, F.ones_like(x), x)

        # calculate likelihood for values which are not NaN
        non_nan_dist_log_likelihood = F.where(
            x != x,
            -x.ones_like() / 0.0,
            self.components[0].log_prob(x_non_nan),
        )

        log_mix_weights = F.log(self.mixture_probs)

        # stack log probabilities of components
        component_log_likelihood = F.stack(
            *[non_nan_dist_log_likelihood, self.components[1].log_prob(x)],
            axis=-1,
        )
        # compute mixture log probability by log-sum-exp
        summands = log_mix_weights + component_log_likelihood
        max_val = F.max_axis(summands, axis=-1, keepdims=True)

        sum_exp = F.sum(
            F.exp(F.broadcast_minus(summands, max_val)), axis=-1, keepdims=True
        )

        log_sum_exp = F.log(sum_exp) + max_val
        return log_sum_exp.squeeze(axis=-1)


class NanMixtureArgs(gluon.HybridBlock):
    def __init__(
        self,
        distr_output: DistributionOutput,
        prefix: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.component_projection: gluon.HybridBlock
        with self.name_scope():
            self.proj_nan_prob = gluon.nn.HybridSequential()
            self.proj_nan_prob.add(
                gluon.nn.Dense(1, prefix=f"{prefix}_pi_", flatten=False)
            )
            self.proj_nan_prob.add(gluon.nn.HybridLambda("sigmoid"))

            self.component_projection = distr_output.get_args_proj()

            self.register_child(self.component_projection)

    def hybrid_forward(self, F, x: Tensor) -> Tuple[Tensor, ...]:
        nan_prob = self.proj_nan_prob(x)
        component_args = self.component_projection(x)
        return tuple([nan_prob.squeeze(axis=-1), component_args])


class NanMixtureOutput(DistributionOutput):
    distr_cls: type = NanMixture

    @validated()
    def __init__(self, distr_output: DistributionOutput) -> None:
        self.distr_output = distr_output

    def get_args_proj(self, prefix: Optional[str] = None) -> NanMixtureArgs:
        return NanMixtureArgs(self.distr_output, prefix=prefix)

    # Overwrites the parent class method.
    def distribution(
        self,
        distr_args,
        loc: Optional[Tensor] = None,
        scale: Optional[Tensor] = None,
        **kwargs,
    ) -> MixtureDistribution:
        nan_prob = distr_args[0]
        component_args = distr_args[1]
        return NanMixture(
            nan_prob=nan_prob,
            distribution=self.distr_output.distribution(
                component_args, loc=loc, scale=scale
            ),
        )

    @property
    def event_shape(self) -> Tuple:
        return self.distr_output.event_shape



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx
import numpy as np

import pytest
from mxnet import nd

from gluonts.mx.model.tpp.deeptpp._network import (
    DeepTPPPredictionNetwork,
    DeepTPPTrainingNetwork,
)
from gluonts.mx.model.tpp.distribution import WeibullOutput


def _allclose(a: nd.NDArray, b: nd.NDArray):
    return np.allclose(a.asnumpy(), b.asnumpy(), atol=1e-6)


TEST_CASES = [  # ia_times, marks, valid_length, num_marks, loglike
    (
        nd.array([[0.1, 0.2, 0.1, 0.12], [0.3, 0.15, 0.1, 0.12]]),
        nd.array([[1, 2, 0, 2], [0, 0, 1, 2]]),
        nd.array([3, 4]),
        3,
        nd.array([5.0570173, 6.6502876]),
    ),
    (
        nd.array([[0.1, 0.2, 0.1, 0.12], [0.3, 0.15, 0.1, 0.12]]),
        nd.array([[1, 2, 0, 2], [0, 0, 1, 2]]),
        nd.array([0, 4]),
        3,
        nd.array([0.9517526, 6.6502876]),
    ),
    (
        nd.array([[0.1, 0.2, 0.1, 0.12], [0.3, 0.15, 0.1, 0.12]]),
        nd.array([[1, 2, 0, 2], [0, 0, 1, 2]]),
        nd.array([0, 0]),
        3,
        nd.array([0.9517526, 0.9725293]),
    ),
    (
        nd.array([[0.1, 0.2, 0.1, 0.12], [0.3, 0.15, 0.1, 0.12]]),
        nd.array([[0, 0, 0, 0], [0, 0, 0, 0]]),
        nd.array([3, 4]),
        1,
        nd.array([1.7608211, 2.2495322]),
    ),
]


@pytest.mark.parametrize(
    "ia_times, marks, valid_length, num_marks, loglike", TEST_CASES
)
def test_log_likelihood(ia_times, marks, valid_length, num_marks, loglike):
    mx.rnd.seed(seed_state=1234)

    model = DeepTPPTrainingNetwork(
        num_marks=num_marks,
        interval_length=2,
        time_distr_output=WeibullOutput(),
    )
    model.initialize()

    loglike_pred = model(nd.stack(ia_times, marks, axis=-1), valid_length)

    assert loglike_pred.shape == (ia_times.shape[0],)
    assert _allclose(loglike, loglike_pred)


def test_trainining_network_disallows_hybrid():
    mx.rnd.seed(seed_state=1234)

    with pytest.raises(NotImplementedError):
        smodel = DeepTPPTrainingNetwork(num_marks=3, interval_length=2)
        smodel.hybridize()


def test_prediction_network_disallows_hybrid():
    mx.rnd.seed(seed_state=1234)

    with pytest.raises(NotImplementedError):
        smodel = DeepTPPPredictionNetwork(
            num_marks=3, interval_length=2, prediction_interval_length=3
        )
        smodel.hybridize()


def test_prediction_network_output():
    mx.rnd.seed(seed_state=1234)
    model = DeepTPPPredictionNetwork(
        num_marks=5,
        time_distr_output=WeibullOutput(),
        interval_length=1.0,
        prediction_interval_length=10.0,
    )
    model.initialize()
    past_ia_times = nd.array([[0.1, 0.2, 0.1, 0.12], [0.3, 0.15, 0.1, 0.12]])
    past_marks = nd.array([[1, 2, 0, 2], [0, 0, 1, 2]])
    past_valid_length = nd.array([3, 4])
    past_target = nd.stack(past_ia_times, past_marks, axis=-1)

    pred_target, pred_valid_length = model(past_target, past_valid_length)

    # pred_target must have shape
    # (num_parallel_samples, batch_size, max_sequence_length, 2)
    assert pred_target.ndim == 4
    assert pred_target.shape[0] == model.num_parallel_samples
    assert pred_target.shape[1] == past_ia_times.shape[0]
    assert pred_target.shape[3] == 2  # TPP prediction contains ia_time & mark
    # pred_valid_length must have shape (num_parallel_samples, batch_size)
    assert pred_valid_length.ndim == 2
    assert pred_valid_length.shape[0] == model.num_parallel_samples
    assert pred_valid_length.shape[1] == past_ia_times.shape[0]

    pred_ia_times = pred_target[..., 0].asnumpy()
    pred_marks = pred_target[..., 1].asnumpy()

    assert pred_marks.min() >= 0
    assert pred_marks.max() < model.num_marks
    assert (pred_ia_times >= 0).all()
    # ia_times are set to zero above valid_length (see DeepTPPPredictionNetwork)
    assert (pred_ia_times.sum(-1) < model.prediction_interval_length).all()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytorch_lightning as pl
import torch

from gluonts.core.component import validated
from gluonts.torch.modules.loss import DistributionLoss, NegativeLogLikelihood

from .module import SimpleFeedForwardModel


class SimpleFeedForwardLightningModule(pl.LightningModule):
    """
    A ``pl.LightningModule`` class that can be used to train a
    ``SimpleFeedForwardModel`` with PyTorch Lightning.

    This is a thin layer around a (wrapped) ``SimpleFeedForwardModel`` object,
    that exposes the methods to evaluate training and validation loss.

    Parameters
    ----------
    model
        ``SimpleFeedForwardModel`` to be trained.
    loss
        Loss function to be used for training,
        default: ``NegativeLogLikelihood()``.
    lr
        Learning rate, default: ``1e-3``.
    weight_decay
        Weight decay regularization parameter, default: ``1e-8``.
    """

    @validated()
    def __init__(
        self,
        model_kwargs: dict,
        loss: DistributionLoss = NegativeLogLikelihood(),
        lr: float = 1e-3,
        weight_decay: float = 1e-8,
    ):
        super().__init__()
        self.save_hyperparameters()
        self.model = SimpleFeedForwardModel(**model_kwargs)
        self.loss = loss
        self.lr = lr
        self.weight_decay = weight_decay

    def forward(self, *args, **kwargs):
        return self.model.forward(*args, **kwargs)

    def _compute_loss(self, batch):
        context = batch["past_target"]
        target = batch["future_target"]
        observed_target = batch["future_observed_values"]

        assert context.shape[-1] == self.model.context_length
        assert target.shape[-1] == self.model.prediction_length

        distr_args, loc, scale = self.model(context)
        distr = self.model.distr_output.distribution(distr_args, loc, scale)

        return (
            self.loss(distr, target) * observed_target
        ).sum() / torch.maximum(torch.tensor(1.0), observed_target.sum())

    def training_step(self, batch, batch_idx: int):  # type: ignore
        """
        Execute training step.
        """
        train_loss = self._compute_loss(batch)
        self.log(
            "train_loss",
            train_loss,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )
        return train_loss

    def validation_step(self, batch, batch_idx: int):  # type: ignore
        """
        Execute validation step.
        """
        val_loss = self._compute_loss(batch)
        self.log(
            "val_loss", val_loss, on_epoch=True, on_step=False, prog_bar=True
        )
        return val_loss

    def configure_optimizers(self):
        """
        Returns the optimizer to use.
        """
        return torch.optim.Adam(
            self.model.parameters(),
            lr=self.lr,
            weight_decay=self.weight_decay,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

try:
    import click
except ImportError:
    import sys

    print("`gluonts.info` requires `click` to be installed.", file=sys.stderr)
    sys.exit(1)

import gluonts


@click.group()
def cli():
    pass


@cli.command()
def version():
    click.echo(gluonts.__version__)


if __name__ == "__main__":
    cli()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np

from gluonts.ev import (
    absolute_error,
    absolute_label,
    absolute_percentage_error,
    coverage,
    error,
    quantile_loss,
    squared_error,
    symmetric_absolute_percentage_error,
    scaled_interval_score,
    absolute_scaled_error,
    seasonal_error,
)

PREDICTION_LENGTH = 5

NAN = np.full(PREDICTION_LENGTH, np.nan)
ZEROES = np.zeros(PREDICTION_LENGTH)
CONSTANT = np.full(PREDICTION_LENGTH, 0.4)
LINEAR = np.array([0.0, 0.1, 0.2, 0.3, 0.4])
EXPONENTIAL = np.array([0.1, 0.01, 0.001, 0.0001, 0.00001])
RANDOM = np.random.random(PREDICTION_LENGTH)
NEGATIVE_RANDOM = -1 * np.random.random(PREDICTION_LENGTH)

TIME_SERIES = [
    NAN,
    ZEROES,
    CONSTANT,
    LINEAR,
    EXPONENTIAL,
    RANDOM,
    NEGATIVE_RANDOM,
]


def test_absolute_label():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            data = {"label": label, "mean": forecast}
            actual = absolute_label(data)
            expected = np.abs(label)

            np.testing.assert_almost_equal(actual, expected)


def test_error():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            data = {"label": label, "0.5": forecast}
            actual = error(data, forecast_type="0.5")
            expected = label - forecast

            np.testing.assert_almost_equal(actual, expected)


def test_abs_error():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            data = {"label": label, "0.5": forecast}
            actual = absolute_error(data, forecast_type="0.5")
            expected = np.abs(label - forecast)

            np.testing.assert_almost_equal(actual, expected)


def test_squared_error():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            data = {"label": label, "mean": forecast}
            actual = squared_error(data, forecast_type="mean")
            expected = np.square(label - forecast)

            np.testing.assert_almost_equal(actual, expected)


def test_quantile_loss():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            q = 0.9
            data = {"label": label, str(q): forecast}
            actual = quantile_loss(data, q=q)
            expected = 2 * np.abs(
                (label - forecast) * ((forecast >= label) - q)
            )

            np.testing.assert_almost_equal(actual, expected)


def test_coverage():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            q = 0.9
            data = {"label": label, str(q): forecast}
            actual = coverage(data, q=q)
            expected = label <= forecast

            np.testing.assert_almost_equal(actual, expected)


def test_absolute_percentage_error():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            data = {"label": label, "0.5": forecast}
            actual = absolute_percentage_error(data, forecast_type="0.5")
            expected = np.abs(label - forecast) / np.abs(label)

            np.testing.assert_almost_equal(actual, expected)


def test_symmetric_absolute_percentage_error():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            data = {"label": label, "0.5": forecast}
            actual = symmetric_absolute_percentage_error(
                data, forecast_type="0.5"
            )
            expected = (
                2
                * np.abs(label - forecast)
                / (np.abs(label) + np.abs(forecast))
            )

            np.testing.assert_almost_equal(actual, expected)


def test_scaled_interval_score():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            alpha = 0.05

            # to keep things simple, the following values are rather arbitrary
            lower_quantile = forecast * (alpha / 2)
            upper_quantile = forecast * (1 - alpha / 2)

            # applying `seasonal_error` on the label is not realistic but
            # at least, the seasonal error function gets used this way
            seasonal_err = seasonal_error(label, seasonality=2)

            data = {
                "label": label,
                "mean": forecast,
                "seasonal_error": seasonal_err,
                "0.025": lower_quantile,
                "0.975": upper_quantile,
            }

            actual = scaled_interval_score(data, alpha=alpha)
            expected = (
                upper_quantile
                - lower_quantile
                + 2.0
                / alpha
                * (lower_quantile - label)
                * (label < lower_quantile)
                + 2.0
                / alpha
                * (label - upper_quantile)
                * (label > upper_quantile)
            ) / seasonal_err

            np.testing.assert_almost_equal(actual, expected)


def test_absolute_scaled_error():
    for label in TIME_SERIES:
        for forecast in TIME_SERIES:
            # applying `seasonal_error` on the label is not realistic but
            # at least, the seasonal error function gets used this way
            seasonal_err = seasonal_error(label, seasonality=2)
            data = {
                "label": label,
                "0.5": forecast,
                "seasonal_error": seasonal_err,
            }

            actual = absolute_scaled_error(data, forecast_type="0.5")
            expected = np.abs(label - forecast) / seasonal_err

            np.testing.assert_almost_equal(actual, expected)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pygments.style import Style
from pygments.token import Comment, Keyword, Name, String, Number, Operator

from . import colors


class Dark(Style):
    # transparent
    background_color = "rgba(0, 0, 0, 0.0)"

    styles = {
        Comment: "italic #888",
        Comment.Preproc: "noitalic #888",
        Keyword: f"bold {colors.SALMON}",
        Keyword.Pseudo: "nobold",
        Keyword.Type: f"nobold {colors.SALMON}",
        Operator: f"bold {colors.SALMON}",
        Operator.Word: f"bold {colors.SALMON}",
        Name.Builtin: colors.YELLOW,
        Name.Function: f"{colors.GREEN}",
        Name.Class: f"bold {colors.GREEN}",
        Name.Namespace: f"bold {colors.GREEN}",
        Name.Variable: colors.YELLOW,
        Name.Constant: colors.RED,
        Name.Label: colors.YELLOW,
        Name.Attribute: colors.YELLOW,
        Name.Tag: f"bold {colors.SALMON}",
        Name.Decorator: colors.SALMON,
        String: colors.SALMON,
        String.Doc: "italic",
        Number: colors.GREEN,
    }



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Type

import mxnet as mx
import numpy as np

from gluonts.mx import Tensor


def batch_diagonal(
    F,
    matrix: Tensor,
    num_data_points: Optional[int] = None,
    float_type=np.float32,
) -> Tensor:
    """
    This function extracts the diagonal of a batch matrix.

    Parameters
    ----------
    F
        A module that can either refer to the Symbol API or the NDArray
        API in MXNet.
    matrix
        matrix of shape (batch_size, num_data_points, num_data_points).
    num_data_points
        Number of rows in the kernel_matrix.

    Returns
    -------
    Tensor
        Diagonals of kernel_matrix of shape (batch_size, num_data_points, 1).
    """
    return F.linalg.gemm2(
        F.broadcast_mul(F.eye(num_data_points, dtype=float_type), matrix),
        F.ones_like(F.slice_axis(matrix, axis=2, begin=0, end=1)),
    )


def lower_triangular_ones(F, d: int, offset: int = 0) -> Tensor:
    """
    Constructs a lower triangular matrix consisting of ones.

    Parameters
    ----------
    F
    d
        Dimension of the output tensor, whose shape will be (d, d).
    offset
        Indicates how many diagonals to set to zero in the lower triangular
        part. By default, offset = 0, so the output matrix contains also the
        main diagonal. For example, if offset = 1 then the output will be a
        strictly lower triangular matrix (i.e. the main diagonal will be zero).

    Returns
    -------
    Tensor
        Tensor of shape (d, d) consisting of ones in the strictly lower
        triangular part, and zeros elsewhere.
    """
    mask = F.zeros_like(F.eye(d))
    for k in range(offset, d):
        mask = mask + F.eye(d, d, -k)
    return mask


def jitter_cholesky_eig(
    F,
    matrix: Tensor,
    num_data_points: Optional[int] = None,
    float_type: Type = np.float64,
    diag_weight: float = 1e-6,
) -> Tensor:
    """
    This function applies the jitter method using the eigenvalue decomposition.
    The eigenvalues are bound below by the jitter, which is proportional to the
    mean of the diagonal elements.

    Parameters
    ----------
    F
        A module that can either refer to the Symbol API or the NDArray
        API in MXNet.
    matrix
        Matrix of shape (batch_size, num_data_points, num_data_points).
    num_data_points
        Number of rows in the kernel_matrix.
    float_type
        Determines whether to use single or double precision.

    Returns
    -------
    Tensor
        Returns the approximate lower triangular Cholesky factor `L`
        of shape (batch_size, num_data_points, num_data_points)
    """
    diag = batch_diagonal(
        F, matrix, num_data_points, float_type
    )  # shape (batch_size, num_data_points, 1)
    diag_mean = diag.mean(axis=1).expand_dims(
        axis=2
    )  # shape (batch_size, 1, 1)
    U, Lambda = F.linalg.syevd(matrix)
    jitter = F.broadcast_mul(diag_mean, F.ones_like(diag)) * diag_weight
    # `K = U^TLambdaU`, where the rows of `U` are the eigenvectors of `K`.
    # The eigendecomposition :math:`U^TLambdaU` is used instead of
    # :math: ULambdaU^T, sine to utilize row-based computation
    # (see Section 4, Seeger et al., 2018)
    return F.linalg.potrf(
        F.linalg.gemm2(
            U,
            F.linalg.gemm2(
                F.broadcast_mul(
                    F.eye(num_data_points, dtype=float_type),
                    F.maximum(jitter, Lambda.expand_dims(axis=2)),
                ),
                U,
            ),
            transpose_a=True,
        )
    )


def jitter_cholesky(
    F,
    matrix: Tensor,
    num_data_points: Optional[int] = None,
    float_type: Type = np.float64,
    max_iter_jitter: int = 10,
    neg_tol: float = -1e-8,
    diag_weight: float = 1e-6,
    increase_jitter: int = 10,
) -> Optional[Tensor]:
    """
    This function applies the jitter method.  It iteratively tries to compute
    the Cholesky decomposition and adds a positive tolerance to the diagonal
    that increases at each iteration until the matrix is positive definite or
    the maximum number of iterations has been reached.

    Parameters
    ----------
    matrix
        Kernel matrix of shape (batch_size, num_data_points, num_data_points).
    num_data_points
        Number of rows in the kernel_matrix.
    float_type
        Determines whether to use single or double precision.
    max_iter_jitter
        Maximum number of iterations for jitter to iteratively make the matrix
        positive definite.
    neg_tol
        Parameter in the jitter methods to eliminate eliminate matrices with
        diagonal elements smaller than this when checking if a matrix is
        positive definite.
    diag_weight
            Multiple of mean of diagonal entries to initialize the jitter.
    increase_jitter
        Each iteration multiply by jitter by this amount
    Returns
    -------
    Optional[Tensor]
        The method either fails to make the matrix positive definite within the
        maximum number of iterations and outputs an error or succeeds and
        returns the lower triangular Cholesky factor `L` of shape
        (batch_size, num_data_points, num_data_points)
    """
    num_iter = 0
    diag = batch_diagonal(
        F, matrix, num_data_points, float_type
    )  # shape (batch_size, num_data_points, 1)
    diag_mean = diag.mean(axis=1).expand_dims(
        axis=2
    )  # shape (batch_size, 1, 1)
    jitter = F.zeros_like(diag)  # shape (batch_size, num_data_points, 1)
    # Ensure that diagonal entries are numerically non-negative, as defined by
    # neg_to
    # TODO: Add support for symbolic case: Cannot use < operator with symbolic
    # variables
    if F.sum(diag <= neg_tol) > 0:
        raise mx.base.MXNetError(
            " Matrix is not positive definite: negative diagonal elements"
        )
    while num_iter <= max_iter_jitter:
        try:
            L = F.linalg.potrf(
                F.broadcast_add(
                    matrix,
                    F.broadcast_mul(
                        F.eye(num_data_points, dtype=float_type),
                        jitter,
                    ),
                )
            )
            # gpu will not throw error but will store nans. If nan, L.sum() =
            # nan and L.nansum() computes the sum treating nans as zeros so
            # the error tolerance can be large. for axis = Null, nansum() and
            # sum() will sum over all elements and return scalar array with
            # shape (1,) TODO: Add support for symbolic case: Cannot use <=
            # operator with symbolic variables
            assert F.abs(L.nansum() - L.sum()) <= 1e-1
            return L
        except Exception:  # TODO: this looks weird
            if num_iter == 0:
                # Initialize the jitter: constant jitter per each batch
                jitter = (
                    F.broadcast_mul(diag_mean, F.ones_like(jitter))
                    * diag_weight
                )
            else:
                jitter = jitter * increase_jitter
        finally:
            num_iter += 1
    raise mx.base.MXNetError(
        " Matrix is not positive definite after the maximum number of"
        f" iterations = {max_iter_jitter} with a maximum jitter ="
        f" {F.max(jitter)}"
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Optional, Tuple, Union, cast

import mxnet as mx
import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.util import cumsum

from .bijection import AffineTransformation, Bijection
from .distribution import Distribution, getF
from .distribution_output import ArgProj, DistributionOutput
from .transformed_distribution import TransformedDistribution


class PiecewiseLinear(Distribution):
    r"""
    Piecewise linear distribution.

    This class represents the *quantile function* (i.e., the inverse CDF)
    associated with the a distribution, as a continuous, non-decreasing,
    piecewise linear function defined in the [0, 1] interval:

    .. math::
        q(x; \gamma, b, d) = \gamma + \sum_{l=0}^L b_l (x_l - d_l)_+

    where the input :math:`x \in [0,1]` and the parameters are

    - :math:`\gamma`: intercept at 0
    - :math:`b`: differences of the slopes in consecutive pieces
    - :math:`d`: knot positions

    Parameters
    ----------
    gamma
        Tensor containing the intercepts at zero
    slopes
        Tensor containing the slopes of each linear piece.
        All coefficients must be positive.
        Shape: ``(*gamma.shape, num_pieces)``
    knot_spacings
        Tensor containing the spacings between knots in the splines.
        All coefficients must be positive and sum to one on the last axis.
        Shape: ``(*gamma.shape, num_pieces)``
    F
    """

    is_reparameterizable = False

    @validated()
    def __init__(
        self, gamma: Tensor, slopes: Tensor, knot_spacings: Tensor
    ) -> None:
        self.gamma = gamma
        self.slopes = slopes
        self.knot_spacings = knot_spacings

        # Since most of the calculations are easily expressed in the original
        # parameters, we transform the learned parameters back
        self.b, self.knot_positions = PiecewiseLinear._to_orig_params(
            self.F, slopes, knot_spacings
        )

    @property
    def F(self):
        return getF(self.gamma)

    @property
    def args(self) -> List:
        return [self.gamma, self.slopes, self.knot_spacings]

    @staticmethod
    def _to_orig_params(
        F, slopes: Tensor, knot_spacings: Tensor
    ) -> Tuple[Tensor, Tensor]:
        r"""
        Convert the trainable parameters to the original parameters of the
        splines, i.e., convert the slopes of each piece to the difference
        between slopes of consecutive pieces and knot spacings to knot
        positions.

        Parameters
        ----------
        F
        slopes
            Tensor of shape (*gamma.shape, num_pieces)
        knot_spacings
            Tensor of shape (*gamma.shape, num_pieces)

        Returns
        -------
        Tensor
            Tensor of shape (*gamma.shape, num_pieces)
        Tensor
            Tensor of shape (*gamma.shape, num_pieces)
        """

        # b: the difference between slopes of consecutive pieces
        # shape (..., num_pieces - 1)
        b = F.slice_axis(slopes, axis=-1, begin=1, end=None) - F.slice_axis(
            slopes, axis=-1, begin=0, end=-1
        )

        # Add slope of first piece to b: b_0 = m_0
        m_0 = F.slice_axis(slopes, axis=-1, begin=0, end=1)
        b = F.concat(m_0, b, dim=-1)

        # The actual position of the knots is obtained by cumulative sum of
        # the knot spacings. The first knot position is always 0 for quantile
        # functions; cumsum will take care of that.
        knot_positions = cumsum(F, knot_spacings, exclusive=True)

        return b, knot_positions

    def sample(
        self, num_samples: Optional[int] = None, dtype=np.float32
    ) -> Tensor:
        F = self.F

        # if num_samples=None then u should have the same shape as gamma,
        # i.e., (dim,) else u should be (num_samples, dim)

        # Note: there is no need to extend the parameters to
        # (num_samples, dim, ...) Thankfully samples returned by
        # `uniform_like` have the expected datatype.
        u = F.random.uniform_like(
            data=(
                self.gamma
                if num_samples is None
                else self.gamma.expand_dims(axis=0).repeat(
                    axis=0, repeats=num_samples
                )
            )
        )

        sample = self.quantile(u)

        if num_samples is None:
            sample = F.squeeze(sample, axis=0)

        return sample

    # overwrites the loss method of the Distribution class
    def loss(self, x: Tensor) -> Tensor:
        return self.crps(x)

    def crps(self, x: Tensor) -> Tensor:
        r"""
        Compute CRPS in analytical form.

        Parameters
        ----------
        x
            Observation to evaluate. Shape equals to gamma.shape.

        Returns
        -------
        Tensor
            Tensor containing the CRPS.
        """

        F = self.F
        gamma, b, knot_positions = self.gamma, self.b, self.knot_positions

        a_tilde = self.cdf(x)

        max_a_tilde_knots = F.broadcast_maximum(
            a_tilde.expand_dims(axis=-1), knot_positions
        )

        knots_cubed = F.power(
            knot_positions, F.ones_like(knot_positions) * 3.0
        )

        coeff = (
            (1.0 - knots_cubed) / 3.0
            - knot_positions
            - F.square(max_a_tilde_knots)
            + 2 * max_a_tilde_knots * knot_positions
        )

        crps = (
            (2 * a_tilde - 1) * x
            + (1 - 2 * a_tilde) * gamma
            + F.sum(b * coeff, axis=-1, keepdims=False)
        )

        return crps

    def cdf(self, x: Tensor) -> Tensor:
        r"""
        Computes the quantile level :math:`\alpha` such that
        :math:`q(\alpha) = x`.

        Parameters
        ----------
        x
            Tensor of shape gamma.shape

        Returns
        -------
        Tensor
            Tensor of shape gamma.shape
        """

        F = self.F
        gamma, b, knot_positions = self.gamma, self.b, self.knot_positions

        quantiles_at_knots = self.quantile_internal(knot_positions, axis=-2)

        # Mask to nullify the terms corresponding to knots larger than l_0,
        # which is the largest knot(quantile level) such that the quantile at
        # l_0, s(l_0) < x.(..., num_pieces)
        mask = F.broadcast_lesser(quantiles_at_knots, x.expand_dims(axis=-1))

        slope_l0 = F.sum(b * mask, axis=-1, keepdims=False)

        # slope_l0 can be zero in which case a_tilde = 0. The following is to
        # circumvent mxnet issue with "where" operator which returns nan even
        # if the statement you are interested in does not result in nan
        # (but the "else" statement evaluates to nan).
        slope_l0_nz = F.where(
            slope_l0 == F.zeros_like(slope_l0), F.ones_like(x), slope_l0
        )

        a_tilde = F.where(
            slope_l0 == F.zeros_like(slope_l0),
            F.zeros_like(x),
            (
                x
                - gamma
                + F.sum(b * knot_positions * mask, axis=-1, keepdims=False)
            )
            / slope_l0_nz,
        )

        return F.broadcast_minimum(F.ones_like(a_tilde), a_tilde)

    def quantile(self, level: Tensor) -> Tensor:
        return self.quantile_internal(level, axis=0)

    def quantile_internal(
        self, x: Tensor, axis: Optional[int] = None
    ) -> Tensor:
        r"""
        Evaluates the quantile function at the quantile levels contained in
        `x`.

        Parameters
        ----------
        x
            Tensor of shape ``*gamma.shape`` if axis=None, or containing an
            additional axis on the specified position, otherwise.
        axis
            Index of the axis containing the different quantile levels which
            are to be computed.

        Returns
        -------
        Tensor
            Quantiles tensor, of the same shape as x.
        """

        F = self.F

        # shapes of self
        # self.gamma: (*batch_shape)
        # self.knot_positions, self.b: (*batch_shape, num_pieces)

        # axis=None - passed at inference when num_samples is None
        # The shape of x is (*batch_shape).
        # The shapes of the parameters should be:
        # gamma: (*batch_shape), knot_positions, b: (*batch_shape, num_pieces)
        # They match the self. counterparts so no reshaping is needed

        # axis=0 - passed at inference when num_samples is not None
        # The shape of x is (num_samples, *batch_shape).
        # The shapes of the parameters should be:
        # gamma: (num_samples, *batch_shape), knot_positions, b:
        # (num_samples, *batch_shape, num_pieces), They do not match the self.
        # counterparts and we need to expand the axis=0 to all of them.

        # axis=-2 - passed at training when we evaluate quantiles at
        # knot_positions in order to compute a_tilde
        # The shape of x is shape(x) = shape(knot_positions) =
        # (*batch_shape, num_pieces).
        # The shape of the parameters shopuld be:
        # gamma: (*batch_shape, 1), knot_positions: (*batch_shape, 1,
        # num_pieces), b: (*batch_shape, 1, num_pieces) They do not match the
        # self. counterparts and we need to expand axis=-1 for gamma and
        # axis=-2 for the rest.

        if axis is not None:
            gamma = self.gamma.expand_dims(axis=axis if axis == 0 else -1)
            knot_positions = self.knot_positions.expand_dims(axis=axis)
            b = self.b.expand_dims(axis=axis)
        else:
            gamma, knot_positions, b = self.gamma, self.knot_positions, self.b

        x_minus_knots = F.broadcast_minus(
            x.expand_dims(axis=-1), knot_positions
        )

        quantile = F.broadcast_add(
            gamma, F.sum(F.broadcast_mul(b, F.relu(x_minus_knots)), axis=-1)
        )

        return quantile

    @property
    def batch_shape(self) -> Tuple:
        return self.gamma.shape

    @property
    def event_shape(self) -> Tuple:
        return ()

    @property
    def event_dim(self) -> int:
        return 0


class PiecewiseLinearOutput(DistributionOutput):
    distr_cls: type = PiecewiseLinear

    @validated()
    def __init__(self, num_pieces: int) -> None:
        super().__init__(self)

        assert (
            isinstance(num_pieces, int) and num_pieces > 1
        ), "num_pieces should be an integer larger than 1"

        self.num_pieces = num_pieces
        self.args_dim = cast(
            Dict[str, int],
            {"gamma": 1, "slopes": num_pieces, "knot_spacings": num_pieces},
        )

    @classmethod
    def domain_map(cls, F, gamma, slopes, knot_spacings):
        # slopes of the pieces are non-negative
        slopes_proj = F.Activation(data=slopes, act_type="softrelu") + 1e-4

        # the spacing between the knots should be in [0, 1] and sum to 1
        knot_spacings_proj = F.softmax(knot_spacings)

        return gamma.squeeze(axis=-1), slopes_proj, knot_spacings_proj

    def distribution(
        self,
        distr_args,
        loc: Optional[Tensor] = None,
        scale: Optional[Tensor] = None,
    ) -> PiecewiseLinear:
        if scale is None:
            return self.distr_cls(*distr_args)
        else:
            distr = self.distr_cls(*distr_args)
            return TransformedPiecewiseLinear(
                distr, [AffineTransformation(loc=loc, scale=scale)]
            )

    @property
    def event_shape(self) -> Tuple:
        return ()


class FixedKnotsArgProj(ArgProj):
    def __init__(self, knot_spacings: mx.nd.NDArray, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_pieces = len(knot_spacings)
        with self.name_scope():
            self.knot_spacings = self.params.get_constant(
                "knot_spacings", knot_spacings
            )

    def hybrid_forward(self, F, x: Tensor, **kwargs) -> Tuple[Tensor]:
        params_unbounded = [proj(x) for proj in self.proj]
        knot_spacings = kwargs["knot_spacings"]

        ks_proj = F.broadcast_add(
            params_unbounded[0].zeros_like(), knot_spacings
        )

        return self.domain_map(*params_unbounded, ks_proj)


class FixedKnotsPiecewiseLinearOutput(PiecewiseLinearOutput):
    """
    A simple extension of PiecewiseLinearOutput that "fixes" the knot positions
    in the quantile function representation. That is, instead of initializing
    with the number of pieces, the quantiles are provided directly at
    initialization.

    Parameters
    ----------
    quantile_levels
        Points along the domain of the quantile function (i.e., in the
        interval [0,1]) where the knots of the piecewise linear approximation
        will be fixed, provided in sorted order (ascending).

        For more information on the piecewise linear quantile function, refer
        to :code:`gluonts.distribution.PiecewiseLinear`.
    """

    distr_cls: type = PiecewiseLinear

    @validated()
    def __init__(
        self,
        quantile_levels: Union[List[float], np.ndarray],
    ) -> None:
        assert all(
            [0 < q < 1 for q in quantile_levels]
        ), "Quantiles must be strictly between 0 and 1."

        assert np.all(np.diff(quantile_levels) > 0), (
            "Quantiles must be in increasing order, with quantile each"
            " specified once"
        )

        super().__init__(
            num_pieces=len(quantile_levels) + 1,
        )

        # store the "knot spacings" instead of quantiles. see PiecewiseLinear
        # for more information
        self.knot_spacings = np.diff(np.r_[0, quantile_levels, 1])
        self.args_dim: Dict[str, int] = {"gamma": 1, "slopes": self.num_pieces}

    def get_args_proj(self, prefix: Optional[str] = None) -> ArgProj:
        return FixedKnotsArgProj(
            knot_spacings=mx.nd.array(self.knot_spacings),
            args_dim=self.args_dim,
            domain_map=mx.gluon.nn.HybridLambda(self.domain_map),
            prefix=prefix,
            dtype=self.dtype,
        )

    @classmethod
    def domain_map(cls, F, gamma, slopes, knot_spacings):
        # we use the super method only to compute intercepts and slopes
        # TODO: computations on knot spacings could be avoided here
        gamma_out, slopes_out, _ = super().domain_map(
            F, gamma, slopes, knot_spacings
        )

        return gamma_out, slopes_out, knot_spacings


# Need to inherit from PiecewiseLinear to get the overwritten loss method.
class TransformedPiecewiseLinear(TransformedDistribution, PiecewiseLinear):
    @validated()
    def __init__(
        self, base_distribution: PiecewiseLinear, transforms: List[Bijection]
    ) -> None:
        super().__init__(base_distribution, transforms)

    def crps(self, y: Tensor) -> Tensor:
        # TODO: use event_shape
        F = getF(y)
        x = y
        scale = 1.0
        for t in self.transforms[::-1]:
            assert isinstance(
                t, AffineTransformation
            ), "Not an AffineTransformation"
            x = t.f_inv(x)
            scale *= t.scale
        p = self.base_distribution.crps(x)
        return F.broadcast_mul(p, scale)
